------------------------------------------------------------------------------
--                                                                          --
--                         GNAT COMPILER COMPONENTS                         --
--                                                                          --
--                              B I N D G E N                               --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                            $Revision: 1.108 + 3.11b patches
--                                                                          --
--          Copyright (C) 1992-1998 Free Software Foundation, Inc.          --
--                                                                          --
-- GNAT is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  GNAT is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNAT;  see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- GNAT was originally developed  by the GNAT team at  New York University. --
-- It is now maintained by Ada Core Technologies Inc (http://www.gnat.com). --
--                                                                          --
------------------------------------------------------------------------------

with ALI;      use ALI;
with Binde;    use Binde;
with Casing;   use Casing;
with Fname;    use Fname;
with Gnatvsn;  use Gnatvsn;
with Hostparm;
with Namet;    use Namet;
with Opt;      use Opt;
with Osint;    use Osint;
with Output;  use Output;
with Types;    use Types;
with Sdefault; use Sdefault;

with GNAT.Heap_Sort_A; use GNAT.Heap_Sort_A;
with GNAT.OS_Lib;

package body Bindgen is

   Statement_Buffer : String (1 .. 1000);
   --  Buffer used for constructing output statements

   Last : Natural := 0;
   --  Last location in Statement_Buffer currently set

   With_Finalization : Boolean := False;
   --  Flag which indicates whether the program uses finalization
   --  (presence of the unit System.Finalization_Implementation)

   With_Tasking : Boolean := False;
   --  Flag which indicates whether the program uses tasking
   --  (presence of the unit System.Tasking.Stages)

   Default_Object_Dir : String_Ptr := Object_Dir_Default_Name;
   Default_Object_Len : Integer := Default_Object_Dir.all'Length;

   Num_Elab_Calls : Nat := 0;
   --  Number of generated calls to elaboration routines

   -----------------------
   -- Local Subprograms --
   -----------------------

   procedure WBI (Info : String) renames Osint.Write_Binder_Info;
   --  Convenient shorthand used throughout

   function ABE_Boolean_Required (U : Unit_Id) return Boolean;
   --  Given a unit id value U, determines if the corresponding unit requires
   --  an access-before-elaboration check variable, i.e. it is a non-predefined
   --  body for which no pragma Elaborate, Elaborate_All or Elaborate_Body is
   --  present, and thus could require ABE checks.

   procedure Resolve_Binder_Options;
   --  Set the value of With_Tasking and With_Finalization

   procedure Gen_Adainit_Ada;
   --  Generates the Adainit procedure (Ada code case)

   procedure Gen_Adainit_C;
   --  Generates the Adainit procedure (C code case)

   procedure Gen_Adafinal_Ada;
   --  Generate the Adafinal procedure (Ada code case)

   procedure Gen_Adafinal_C;
   --  Generate the Adafinal procedure (C code case)

   procedure Gen_Elab_Calls_Ada;
   --  Generate sequence of elaboration calls (Ada code case)

   procedure Gen_Elab_Calls_C;
   --  Generate sequence of elaboration calls (C code case)

   procedure Gen_Exception_Table_Ada;
   --  Generate binder exception table (Ada code case). This consists of
   --  declarations followed by a begin followed by a call. If zero cost
   --  exceptions are not active, then only the begin is generated.

   procedure Gen_Exception_Table_C;
   --  Generate binder exception table (C code case). This has no effect
   --  if zero cost exceptions are not active, otherwise it generates a
   --  set of declarations followed by a call.

   procedure Gen_Main_Ada;
   --  Generate procedure main (Ada code case)

   procedure Gen_Main_C;
   --  Generate main() procedure (C code case)

   procedure Gen_Object_Files_Options;
   --  Output comments containing a list of the full names of the object
   --  files to be linked and the list of linker options supplised by
   --  Linker_Options pragmas in the source. (C and Ada code case)

   procedure Gen_Output_File_Ada (Filename : String);
   --  Generate output file (Ada code case)

   procedure Gen_Output_File_C (Filename : String);
   --  Generate output file (C code case)

   procedure Gen_Versions_Ada;
   --  Output series of definitions for unit versions (Ada code case)

   procedure Gen_Versions_C;
   --  Output series of definitions for unit versions (C code case)

   function Lt_Linker_Option (Op1, Op2 : Natural) return Boolean;
   --  Compare linker options, when sorting, first according to
   --  Is_Internal_File (internal files come later) and then by elaboration
   --  order position (latest to earliest) except its not possible to
   --  distinguish between a linker option in the spec and one in the body.

   procedure Move_Linker_Option (From : Natural; To : Natural);
   --  Move routine for sorting linker options

   procedure Public_Version_Warning;
   --  Emit a warning concerning the use of the Public version under
   --  certain circumstances. See details in body.

   procedure Set_Char (C : Character);
   --  Set given character in Statement_Buffer at the Last + 1 position
   --  and increment Last by one to reflect the stored character.

   procedure Set_Int (N : Int);
   --  Set given value in decimal in Statement_Buffer with no spaces
   --  starting at the Last + 1 position, and updating Last past the value.
   --  A minus sign is output for a negative value.

   procedure Set_String (S : String);
   --  Sets characters of given string in Statement_Buffer, starting at the
   --  Last + 1 position, and updating last past the string value.

   procedure Set_Unit_Name;
   --  Given a unit name in the Name_Buffer, copies it to Statement_Buffer,
   --  starting at the Last + 1 position, and updating last past the value.
   --  changing periods to double underscores, and updating Last appropriately.

   procedure Tab_To (N : Natural);
   --  If Last is greater than or equal to N, no effect, otherwise store
   --  blanks in Statement_Buffer bumping Last, until Last = N.

   procedure Write_Info_Ada_C (Ada : String; C : String; Common : String);
   --  For C code case, write C & Common, for Ada case write Ada & Common
   --  to current binder output file using Write_Binder_Info.

   procedure Write_Statement_Buffer;
   --  Write out contents of statement buffer up to Last, and reset Last to 0

   --------------------------
   -- ABE_Boolean_Required --
   --------------------------

   procedure Resolve_Binder_Options is
   begin
      for E in Elab_Order.First .. Elab_Order.Last loop
         Get_Name_String (Unit.Table (Elab_Order.Table (E)).Uname);

         --  if the program uses finalization we must make sure to finalize
         --  global objects too at the end of the program.

         if Name_Buffer (1 .. 34) = "system.finalization_implementation" then
            With_Finalization := True;
         elsif Name_Buffer (1 .. 21) = "system.tasking.stages" then
            With_Tasking := True;
         end if;
      end loop;
   end Resolve_Binder_Options;

   --------------------------
   -- ABE_Boolean_Required --
   --------------------------

   function ABE_Boolean_Required (U : Unit_Id) return Boolean is
      Typ   : constant Unit_Type := Unit.Table (U).Utype;
      Units : Unit_Id;

   begin
      if Typ /= Is_Body then
         return False;

      else
         Units := U + 1;

         return (not Unit.Table (Units).Pure)
                   and then
                (not Unit.Table (Units).Preelab)
                   and then
                (not Unit.Table (Units).Elaborate_Body)
                   and then
                (not Unit.Table (Units).Predefined);
      end if;
   end ABE_Boolean_Required;

   ---------------------
   -- Gen_Adainit_Ada --
   ---------------------

   procedure Gen_Adainit_Ada is
   begin
      --  The flag Ada_Init_Flag is used to ensure that only the first
      --  call to adainit has an effect (RM B.1(39)). It is false (zero)
      --  before the first call, and true thereafter.

      if not Bind_Main_Program then
         WBI ("   Ada_Init_Flag : Boolean := False;");
      end if;

      WBI ("   procedure adainit is");
      WBI ("");
      WBI ("      procedure Set_Globals");
      WBI ("        (Main_Priority            : Integer;");
      WBI ("         Time_Slice_Value         : Integer;");
      WBI ("         WC_Encoding              : Character;");
      WBI ("         Locking_Policy           : Character;");
      WBI ("         Queuing_Policy           : Character;");
      WBI ("         Task_Dispatching_Policy  : Character;");
      WBI ("         Adafinal                 : System.Address;");
      WBI ("         Unreserve_All_Interrupts : Boolean);");
      WBI ("      pragma Import (C, Set_Globals, ""__gnat_set_globals"");");
      WBI ("");

      Gen_Exception_Table_Ada;

      if not Bind_Main_Program then
         WBI ("   if (Ada_Init_Flag) then return; end if");
         WBI ("   Ada_Init_Flag := True;");
      end if;

      WBI ("      Set_Globals");

      Set_String ("        (Main_Priority            => ");
      Set_Int    (ALIs.Table (ALIs.First).Main_Priority);
      Set_Char   (',');
      Write_Statement_Buffer;

      Set_String ("         Time_Slice_Value         => ");
      Set_Int    (ALIs.Table (ALIs.First).Time_Slice_Value);
      Set_Char   (',');
      Write_Statement_Buffer;

      Set_String ("         WC_Encoding              => '");
      Set_Char   (ALIs.Table (ALIs.First).WC_Encoding);
      Set_String ("',");
      Write_Statement_Buffer;

      Set_String ("         Locking_Policy           => '");
      Set_Char   (Locking_Policy);
      Set_String ("',");
      Write_Statement_Buffer;

      Set_String ("         Queuing_Policy           => '");
      Set_Char   (Queuing_Policy);
      Set_String ("',");
      Write_Statement_Buffer;

      Set_String ("         Task_Dispatching_Policy  => '");
      Set_Char   (Task_Dispatching_Policy);
      Set_String ("',");
      Write_Statement_Buffer;

      WBI ("         Adafinal                 => adafinal'Address,");

      Set_String ("         Unreserve_All_Interrupts => ");

      if Unreserve_All_Interrupts then
         Set_String ("True");
      else
         Set_String ("False");
      end if;

      Set_String (");");
      Write_Statement_Buffer;

      Gen_Elab_Calls_Ada;

      WBI ("");
      WBI ("   end adainit;");
   end Gen_Adainit_Ada;

   -------------------
   -- Gen_Adainit_C --
   --------------------

   procedure Gen_Adainit_C is
   begin
      WBI ("void adafinal ();");
      WBI ("");

      WBI ("void adainit ()");
      WBI ("{");

      --  The flag ada__init_flag is used to ensure that only the first
      --  call to adainit has an effect (RM B.1(39)). It is false (zero)
      --  before the first call, and true thereafter.

      if not Bind_Main_Program then
         WBI ("   static int ada__init_flag = 0;");
      end if;

      Gen_Exception_Table_C;

      if not Bind_Main_Program then
         WBI ("   if (ada__init_flag) return;");
         WBI ("   ada__init_flag++;");
      end if;

      --  Generate call to set the runtime global variables defined in
      --  a-init.c. We define the varables in a-init.c, rather than in
      --  the binder generated file itself to avoid undefined externals
      --  when the runtime is linked as a shareable image library.

      --  We call the routine from inside adainit() because this works for
      --  both programs with and those without binder generated "main"
      --  functions. This means, of course, that adafinal() must be defined
      --  before adainit(), so we can pass its address.

      WBI ("   __gnat_set_globals (");

      Set_String ("      ");
      Set_Int (ALIs.Table (ALIs.First).Main_Priority);
      Set_Char (',');
      Tab_To (15);
      Set_String ("/* Main_Priority              */");
      Write_Statement_Buffer;

      Set_String ("      ");
      Set_Int    (ALIs.Table (ALIs.First).Time_Slice_Value);
      Set_Char   (',');
      Tab_To (15);
      Set_String ("/* Time_Slice_Value           */");
      Write_Statement_Buffer;

      Set_String ("      '");
      Set_Char   (ALIs.Table (ALIs.First).WC_Encoding);
      Set_String ("',");
      Tab_To (15);
      Set_String ("/* WC_Encoding                */");
      Write_Statement_Buffer;

      Set_String ("      '");
      Set_Char (Locking_Policy);
      Set_String ("',");
      Tab_To (15);
      Set_String ("/* Locking_Policy             */");
      Write_Statement_Buffer;

      Set_String ("      '");
      Set_Char (Queuing_Policy);
      Set_String ("',");
      Tab_To (15);
      Set_String ("/* Queuing_Policy             */");
      Write_Statement_Buffer;

      Set_String ("      '");
      Set_Char (Task_Dispatching_Policy);
      Set_String ("',");
      Tab_To (15);
      Set_String ("/* Tasking_Dispatching_Policy */");
      Write_Statement_Buffer;

      Set_String ("      ");
      Set_String ("adafinal,");
      Tab_To (15);
      Set_String ("/* Finalization routine address */");
      Write_Statement_Buffer;

      Set_String ("      ");
      Set_Int    (Boolean'Pos (Unreserve_All_Interrupts));
      Set_String (");");
      Tab_To (15);
      Set_String ("/* Unreserve_All_Interrupts */");
      Write_Statement_Buffer;

      WBI ("");
      Gen_Elab_Calls_C;
      WBI ("}");
   end Gen_Adainit_C;

   ----------------------
   -- Gen_Adafinal_Ada --
   ----------------------

   procedure Gen_Adafinal_Ada is
   begin
      WBI ("");
      WBI ("   procedure adafinal is");

      if With_Tasking or With_Finalization then

         WBI ("");
         WBI ("      procedure do_finalize;");
         WBI ("      pragma Import");
         WBI ("        (C, do_finalize,");

         if With_Tasking then
            WBI
              ("         " &
               """system__tasking__stages__finalize_global_tasks"");");

         elsif With_Finalization then
            WBI
              ("         " &
               """system__finalization_implementation" &
               "__finalize_global_list"");");
         end if;

         WBI ("");
         WBI ("   begin");
         WBI ("      do_finalize;");

      else
         WBI ("   begin");
         WBI ("      null;");
      end if;

      WBI ("   end adafinal;");
   end Gen_Adafinal_Ada;

   --------------------
   -- Gen_Adafinal_C --
   --------------------

   procedure Gen_Adafinal_C is
   begin
      WBI ("void adafinal () {");

      if With_Tasking then
         WBI ("   " &
              "system__tasking__stages__finalize_global_tasks ();");

      elsif With_Finalization then
         WBI ("   " &
              "system__finalization_implementation__finalize_global_list ();");
      end if;

      WBI ("}");
   end Gen_Adafinal_C;

   ------------------------
   -- Gen_Elab_Calls_Ada --
   ------------------------

   procedure Gen_Elab_Calls_Ada is
   begin
      for E in Elab_Order.First .. Elab_Order.Last loop
         Get_Name_String (Unit.Table (Elab_Order.Table (E)).Uname);

         --  Generate elaboration call if elaboration needed, and a comment
         --  if no elaboration is required. Note that passive units are
         --  always excluded from elaboration.

         if Unit.Table (Elab_Order.Table (E)).No_Elab
           or else Unit.Table (Elab_Order.Table (E)).Shared_Passive
         then
            Set_String ("      -- ");
         else
            Set_String ("         ");
         end if;

         if Name_Buffer (Name_Len) = 's' then
            Name_Buffer (Name_Len - 1 .. Name_Len + 8) := "'elab_spec";
         else
            Name_Buffer (Name_Len - 1 .. Name_Len + 8) := "'elab_body";
         end if;

         Name_Len := Name_Len + 8;
         Set_Casing (Unit.Table (Elab_Order.Table (E)).Icasing);

         for J in 1 .. Name_Len loop
            Set_Char (Name_Buffer (J));
         end loop;

         Set_Char (';');
         Write_Statement_Buffer;
      end loop;
   end Gen_Elab_Calls_Ada;

   ----------------------
   -- Gen_Elab_Calls_C --
   ----------------------

   procedure Gen_Elab_Calls_C is
   begin
      for E in Elab_Order.First .. Elab_Order.Last loop
         Get_Name_String (Unit.Table (Elab_Order.Table (E)).Uname);

         --  Generate elaboration call if elaboration needed, and a comment
         --  if no elaboration is required. Note that passive units are
         --  always excluded from elaboration.

         if Unit.Table (Elab_Order.Table (E)).No_Elab
           or else Unit.Table (Elab_Order.Table (E)).Shared_Passive
         then
            Set_String ("/* ");
         else
            Set_String ("   ");
         end if;

         Set_Unit_Name;
         Set_String ("___elab");
         Set_Char (Name_Buffer (Name_Len)); -- 's' or 'b' for spec/body
         Set_String (" ();");

         if Unit.Table (Elab_Order.Table (E)).No_Elab
           or else Unit.Table (Elab_Order.Table (E)).Shared_Passive
         then
            Set_String (" */");
         end if;

         Write_Statement_Buffer;
      end loop;
   end Gen_Elab_Calls_C;

   -----------------------------
   -- Gen_Exception_Table_Ada --
   -----------------------------

   procedure Gen_Exception_Table_Ada is
      Num  : Nat;
      Last : ALI_Id;

   begin
      if not Zero_Cost_Exceptions then
         WBI ("   begin");
         return;
      end if;

      --  The code we generate looks like

      --        procedure SDP_Table_Build
      --          (SDP_Addresses   : System.Address;
      --           SDP_Count       : Natural;
      --           Elab_Addresses  : System.Address;
      --           Elab_Addr_Count : Natural);
      --        pragma Import (C, SDP_Table_Build, "__gnat_SDP_Table_Build");
      --
      --        ST : aliased constant array (1 .. nnn) of System.Address := (
      --               unit_name_1'UET_Address,
      --               unit_name_2'UET_Address,
      --               ...
      --               unit_name_3'UET_Address,
      --
      --        EA : aliased constant array (1 .. eee) of System.Address := (
      --               adainit'Code_Address,
      --               adafinal'Code_Address,
      --               unit_name'elab[spec|body]'Address,
      --               unit_name'elab[spec|body]'Address,
      --               unit_name'elab[spec|body]'Address,
      --               unit_name'elab[spec|body]'Address);
      --
      --     begin
      --        SDP_Table_Build (ST'Address, nnn, EA'Address, eee);

      Num := 0;
      for A in ALIs.First .. ALIs.Last loop
         if ALIs.Table (A).Unit_Exception_Table then
            Num := Num + 1;
            Last := A;
         end if;
      end loop;

      WBI ("      procedure SDP_Table_Build");
      WBI ("        (SDP_Addresses   : System.Address;");
      WBI ("         SDP_Count       : Natural;");
      WBI ("         Elab_Addresses  : System.Address;");
      WBI ("         Elab_Addr_Count : Natural);");
      WBI ("      " &
           "pragma Import (C, SDP_Table_Build, ""__gnat_SDP_Table_Build"");");

      WBI (" ");
      Set_String ("      ST : aliased constant array (1 .. ");
      Set_Int (Num);
      Set_String (") of System.Address := (");

      if Num = 1 then
         Set_String ("1 => A1);");
         Write_Statement_Buffer;

      else
         Write_Statement_Buffer;

         for A in ALIs.First .. ALIs.Last loop
            if ALIs.Table (A).Unit_Exception_Table then
               Get_Name_String
                 (Unit.Table (ALIs.Table (A).First_Unit).Uname);
               Set_Casing (Mixed_Case);
               Set_String ("        ");
               Set_String (Name_Buffer (1 .. Name_Len - 2));
               Set_String ("'UET_Address");

               if A = Last then
                  Set_String (");");
               else
                  Set_Char (',');
               end if;

               Write_Statement_Buffer;
            end if;
         end loop;
      end if;

      WBI (" ");
      Set_String ("      EA : aliased constant array (1 .. ");
      Set_Int (Num_Elab_Calls + 2);
      Set_String (") of System.Address := (");
      Write_Statement_Buffer;
      WBI ("        adainit'Code_Address,");

      Set_String ("        adafinal'Code_Address");

      for E in Elab_Order.First .. Elab_Order.Last loop
         Get_Name_String (Unit.Table (Elab_Order.Table (E)).Uname);

         if Unit.Table (Elab_Order.Table (E)).No_Elab
           or else Unit.Table (Elab_Order.Table (E)).Shared_Passive
         then
            null;

         else
            Set_Char (',');
            Write_Statement_Buffer;
            Set_String ("        ");

            if Name_Buffer (Name_Len) = 's' then
               Name_Buffer (Name_Len - 1 .. Name_Len + 16) :=
                                        "'elab_spec'address";
            else
               Name_Buffer (Name_Len - 1 .. Name_Len + 16) :=
                                        "'elab_body'address";
            end if;

            Name_Len := Name_Len + 16;
            Set_Casing (Unit.Table (Elab_Order.Table (E)).Icasing);

            for J in 1 .. Name_Len loop
               Set_Char (Name_Buffer (J));
            end loop;
         end if;
      end loop;

      Set_String (");");
      Write_Statement_Buffer;

      WBI (" ");
      WBI ("   begin");

      Set_String ("      SDP_Table_Build (ST'Address, ");
      Set_Int (Num);
      Set_String (", EA'Address, ");
      Set_Int (Num_Elab_Calls + 2);
      Set_String (");");
      Write_Statement_Buffer;
   end Gen_Exception_Table_Ada;

   ---------------------------
   -- Gen_Exception_Table_C --
   ---------------------------

   procedure Gen_Exception_Table_C is
      Num  : Nat;
      Num2 : Nat;

   begin
      if not Zero_Cost_Exceptions then
         return;
      end if;

      --  The code we generate looks like

      --     extern void *__gnat_unitname1__SDP;
      --     extern void *__gnat_unitname2__SDP;
      --     ...
      --
      --     void **st[nnn] = {
      --       &__gnat_unitname1__SDP,
      --       &__gnat_unitname2__SDP,
      --       ...
      --       &__gnat_unitnamen__SDP};
      --
      --     extern void unitname1__elabb ();
      --     extern void unitname2__elabb ();
      --     ...
      --
      --     void (*ea[eee]) () = {
      --       adainit,
      --       adafinal,
      --       unitname1___elab[b,s],
      --       unitname2___elab[b,s],
      --       ...
      --       unitnamen___elab[b,s]};
      --
      --     __gnat_SDP_Table_Build (&st, nnn, &ea, eee);


      Num := 0;
      for A in ALIs.First .. ALIs.Last loop
         if ALIs.Table (A).Unit_Exception_Table then
            Num := Num + 1;

            Set_String ("   extern void *__gnat_");
            Get_Name_String (Unit.Table (ALIs.Table (A).First_Unit).Uname);
            Set_Unit_Name;
            Set_String ("__SDP");
            Set_Char (';');
            Write_Statement_Buffer;
         end if;
      end loop;

      WBI (" ");

      Set_String ("   void **st[");
      Set_Int (Num);
      Set_String ("] = {");
      Write_Statement_Buffer;

      Num2 := 0;
      for A in ALIs.First .. ALIs.Last loop
         if ALIs.Table (A).Unit_Exception_Table then
            Num2 := Num2 + 1;

            Set_String ("     &__gnat_");
            Get_Name_String (Unit.Table (ALIs.Table (A).First_Unit).Uname);
            Set_Unit_Name;
            Set_String ("__SDP");

            if Num = Num2 then
               Set_String ("};");
            else
               Set_Char (',');
            end if;

            Write_Statement_Buffer;
         end if;
      end loop;

      WBI ("");
      for E in Elab_Order.First .. Elab_Order.Last loop
         Get_Name_String (Unit.Table (Elab_Order.Table (E)).Uname);

         if Unit.Table (Elab_Order.Table (E)).No_Elab
           or else Unit.Table (Elab_Order.Table (E)).Shared_Passive
         then
            null;

         else
            Set_String ("   extern void ");
            Set_Unit_Name;
            Set_String ("___elab");
            Set_Char (Name_Buffer (Name_Len)); -- 's' or 'b' for spec/body
            Set_String (" ();");
            Write_Statement_Buffer;
         end if;
      end loop;

      WBI ("");
      Set_String ("   void (*ea[");
      Set_Int (Num_Elab_Calls + 2);
      Set_String ("]) () = {");
      Write_Statement_Buffer;

      WBI ("     adainit,");
      Set_String ("     adafinal");

      for E in Elab_Order.First .. Elab_Order.Last loop
         Get_Name_String (Unit.Table (Elab_Order.Table (E)).Uname);

         if Unit.Table (Elab_Order.Table (E)).No_Elab
           or else Unit.Table (Elab_Order.Table (E)).Shared_Passive
         then
            null;

         else
            Set_Char (',');
            Write_Statement_Buffer;
            Set_String ("     ");
            Set_Unit_Name;
            Set_String ("___elab");
            Set_Char (Name_Buffer (Name_Len)); -- 's' or 'b' for spec/body
         end if;
      end loop;

      Set_String ("};");
      Write_Statement_Buffer;

      WBI (" ");

      Set_String ("   __gnat_SDP_Table_Build (&st, ");
      Set_Int (Num);
      Set_String (", ea, ");
      Set_Int (Num_Elab_Calls + 2);
      Set_String (");");
      Write_Statement_Buffer;
   end Gen_Exception_Table_C;

   ------------------
   -- Gen_Main_Ada --
   ------------------

   procedure Gen_Main_Ada is
   begin
      WBI ("");
      WBI ("   function main");
      WBI ("     (argc : Integer;");
      WBI ("      argv : System.Address;");
      WBI ("      envp : System.Address)");
      WBI ("      return Integer");
      WBI ("   is");
      WBI ("      procedure initialize;");
      WBI ("      pragma Import (C, initialize, ""__gnat_initialize"");");
      WBI ("");
      WBI ("      procedure finalize;");
      WBI ("      pragma Import (C, finalize, ""__gnat_finalize"");");
      WBI ("");
      WBI ("      procedure Break_Start;");
      WBI ("      pragma Import (C, Break_Start, ""__gnat_break_start"");");

      if ALIs.Table (ALIs.First).Main_Program = Func then
         WBI ("");
         WBI ("      Result : Integer;");
      end if;

      WBI ("");
      WBI ("   begin");
      WBI ("      gnat_argc := argc;");
      WBI ("      gnat_argv := argv;");
      WBI ("      gnat_envp := envp;");
      WBI ("");
      WBI ("      initialize;");
      WBI ("      adainit;");
      WBI ("      Break_Start;");

      Get_Name_String (Unit.Table (First_Unit_Entry).Uname);
      Name_Len := Name_Len - 2;
      Set_Casing (Unit.Table (First_Unit_Entry).Icasing);

      if ALIs.Table (ALIs.First).Main_Program = Proc then
         WBI ("      " & Name_Buffer (1 .. Name_Len) & ';');
         WBI ("      adafinal;");
         WBI ("      finalize;");
         WBI ("      return (gnat_exit_status);");

      else -- function case
         WBI ("      Result := Integer (" &
              Name_Buffer (1 .. Name_Len) & ");");
         WBI ("      adafinal;");
         WBI ("      finalize;");
         WBI ("      return (Result);");
      end if;

      WBI ("   end;");
   end Gen_Main_Ada;

   ----------------
   -- Gen_Main_C --
   ----------------

   procedure Gen_Main_C is

      procedure Set_Main_Program_Name;
      --  Given the main program name in Name_Buffer (length in Name_Len)
      --  generate the name of the routine to be used in the call. The name
      --  is generated starting at Last + 1, and Last is updated past it.

      procedure Set_Main_Program_Name is
      begin
         --  Note that name has %b on the end which we ignore

         --  Output initial _ada_ if no dots in name

         for J in 1 .. Name_Len - 1 loop
            if J = Name_Len - 1 then
               Set_String ("_ada_");
            else
               exit when Name_Buffer (J) = '.';
            end if;
         end loop;

         --  Copy name, changing dots to double underscores

         for J in 1 .. Name_Len - 2 loop
            if Name_Buffer (J) = '.' then
               Set_String ("__");
            else
               Set_Char (Name_Buffer (J));
            end if;
         end loop;
      end Set_Main_Program_Name;

   --  Start of processing for Gen_Main_C

   begin
      if Bind_Alternate_Main_Name then
         WBI ("int gnat_main (argc, argv, envp)");
      else
         WBI ("int main (argc, argv, envp)");
      end if;

      WBI ("    int argc;");
      WBI ("    char **argv;");
      WBI ("    char **envp;");
      WBI ("{");
      WBI ("   gnat_argc = argc;");
      WBI ("   gnat_argv = argv;");
      WBI ("   gnat_envp = envp;");
      WBI (" ");

      WBI ("   __gnat_initialize();");
      WBI ("   adainit();");
      WBI ("   __gnat_break_start();");
      WBI (" ");

      --  Output main program name

      Get_Name_String (Unit.Table (First_Unit_Entry).Uname);

      --  Main program is procedure case

      if ALIs.Table (ALIs.First).Main_Program = Proc then
         Set_String ("   ");
         Set_Main_Program_Name;
         Set_String (" ();");
         Write_Statement_Buffer;

      --  Main program is function case

      else -- ALIs.Table (ALIs_First).Main_Program = Func
         Set_String ("   return (");
         Set_Main_Program_Name;
         Set_String (" ());");
         Write_Statement_Buffer;
      end if;

      WBI (" ");
      WBI ("   adafinal();");
      WBI ("   __gnat_finalize();");

      WBI ("   exit (gnat_exit_status);");
      WBI ("}");
   end Gen_Main_C;

   ------------------------------
   -- Gen_Object_Files_Options --
   ------------------------------

   procedure Gen_Object_Files_Options is
      Lgnat : Integer;

      Wrote_Info_IDENTIFICATION : Boolean := False;

      procedure Write_Linker_Option;
      --  Write binder info linker option, removing duplicated
      --  IDENTIFICATION options.

      procedure Write_Linker_Option is
      begin
         --  Special processing for IDENTIFICATION, remove duplicates

         if Name_Len > 28
           and then Name_Buffer (1 .. 28) = "--for-linker=IDENTIFICATION="
         then
            if not Wrote_Info_IDENTIFICATION then
               Write_Info_Ada_C
                 ("   --   ", "", Name_Buffer (1 .. Name_Len - 1));
               Wrote_Info_IDENTIFICATION := True;
            end if;

         else
            --  Normal case, note that Name_Buffer is null-terminated

            Write_Info_Ada_C ("   --   ", "",
              Name_Buffer (1 .. Name_Len - 1));
         end if;
      end Write_Linker_Option;

   --  Start of processing for Gen_Object_Files_Options

   begin
      WBI ("");
      Write_Info_Ada_C ("--", "/*", " BEGIN Object file/option list");

      for E in Elab_Order.First .. Elab_Order.Last loop

         --  If not spec that has an associated body, then generate a
         --  comment giving the name of the corresponding object file,
         --  except that we always skip shared passive units.

         if Unit.Table (Elab_Order.Table (E)).Utype /= Is_Spec
           and then not Unit.Table (Elab_Order.Table (E)).Shared_Passive
         then
            Get_Name_String
              (ALIs.Table
                (Unit.Table (Elab_Order.Table (E)).My_ALI).Ofile_Full_Name);

            --  If the presence of an object file is necessary or if it
            --  exists, then use it.

            if not Hostparm.Exclude_Missing_Objects
               or else
                 GNAT.OS_Lib.Is_Regular_File (Name_Buffer (1 .. Name_Len))
            then
               Write_Info_Ada_C ("   --   ", "", Name_Buffer (1 .. Name_Len));
            end if;
         end if;
      end loop;

      --  Add a "-Ldir" for each directory in the object path:

      for J in 1 .. Nb_Dir_In_Obj_Search_Path loop
         declare
            Dir : String_Ptr := Dir_In_Obj_Search_Path (J);
            Str : String (1 .. Dir'Length + 2);

         begin
            Str (1 .. 2)        := "-L";
            Str (3 .. Str'Last) := Dir.all;
            Write_Info_Ada_C ("   --   ", "", Str);
         end;
      end loop;

      --  Sort linker options

      Sort (Linker_Options.Last, Move_Linker_Option'Access,
                                 Lt_Linker_Option'Access);

      --  Write user linker options

      Lgnat := Linker_Options.Last + 1;

      for J in 1 .. Linker_Options.Last loop
         if not Linker_Options.Table (J).Internal_File then
            Get_Name_String (Linker_Options.Table (J).Name);
            Write_Linker_Option;
         else
            Lgnat := J;
            exit;
         end if;
      end loop;

      --  Add libgnat.a here since it references symbols in the thread
      --  library which is included as part of the linker options of the
      --  runtime

      Write_Info_Ada_C ("   --   ", "", "-lgnat");

      --  Write internal linker options

      for J in Lgnat .. Linker_Options.Last loop
         Get_Name_String (Linker_Options.Table (J).Name);
         Write_Linker_Option;
      end loop;

      if Ada_Bind_File then
         WBI ("-- END Object file/option list   ");
      else
         WBI ("   END Object file/option list */");
      end if;
   end Gen_Object_Files_Options;

   -------------------------
   -- Gen_Output_File_Ada --
   -------------------------

   procedure Gen_Output_File_Ada (Filename : String) is

      Bfiles : Name_Id;
      --  Name of generated bind file (spec)

      Bfileb : Name_Id;
      --  Name of generated bind file (body)

   begin
      --  Create spec first

      Create_Binder_Output (Filename, 's', Bfiles);

      if Bind_Main_Program then
         WBI ("with System;");
      end if;

      Resolve_Binder_Options;

      WBI ("with Ada.Exceptions;");

      WBI ("package Ada_Main is");

      if Bind_Main_Program then
         WBI ("");
         WBI ("   gnat_argc : Integer;");
         WBI ("   pragma Import (C, gnat_argc);");

         WBI ("");
         WBI ("   gnat_argv : System.Address;");
         WBI ("   pragma Import (C, gnat_argv);");

         WBI ("");
         WBI ("   gnat_envp : System.Address;");
         WBI ("   pragma Import (C, gnat_envp);");

         WBI ("");
         WBI ("   gnat_exit_status : Integer;");
         WBI ("   pragma Import (C, gnat_exit_status);");
      end if;

      WBI ("");
      WBI ("   procedure adafinal;");
      WBI ("   pragma Export (C, adafinal);");

      WBI ("");
      WBI ("   procedure adainit;");
      WBI ("   pragma Export (C, adainit);");

      if Bind_Main_Program then
         WBI ("");
         WBI ("   function main");
         WBI ("     (argc : Integer;");
         WBI ("      argv : System.Address;");
         WBI ("      envp : System.Address)");
         WBI ("      return Integer;");
         WBI ("   pragma Export (C, main, ""main"");");
      end if;

      Gen_Versions_Ada;

      --  Spec is complete

      WBI ("");
      WBI ("end Ada_Main;");
      Close_Binder_Output;

      --  Prepare to write body

      Create_Binder_Output (Filename, 'b', Bfileb);

      --  Output Source_File_Name pragmas which look like

      --    pragma Source_File_Name (Ada_Main, Spec_File_Name => "sss");
      --    pragma Source_File_Name (Ada_Main, Body_File_Name => "bbb");

      --  where sss/bbb are the spec/body file names respectively

      Get_Name_String (Bfiles);
      Name_Buffer (Name_Len + 1 .. Name_Len + 3) := """);";

      WBI ("pragma Source_File_Name (Ada_Main, Spec_File_Name => """ &
           Name_Buffer (1 .. Name_Len + 3));

      Get_Name_String (Bfileb);
      Name_Buffer (Name_Len + 1 .. Name_Len + 3) := """);";

      WBI ("pragma Source_File_Name (Ada_Main, Body_File_Name => """ &
           Name_Buffer (1 .. Name_Len + 3));

      WBI ("");
      Get_Name_String (Unit.Table (First_Unit_Entry).Uname);
      Name_Len := Name_Len - 2;
      Set_Casing (Unit.Table (First_Unit_Entry).Icasing);
      WBI ("with " & Name_Buffer (1 .. Name_Len) & ';');

      WBI ("");
      WBI ("package body Ada_Main is");

      Gen_Adainit_Ada;
      Gen_Adafinal_Ada;

      --  Main is only present for Ada main case

      if Bind_Main_Program then
         Gen_Main_Ada;
      end if;

      --  Output object file list and the Ada body is complete

      Gen_Object_Files_Options;
      WBI ("");
      WBI ("end Ada_Main;");
      Close_Binder_Output;
   end Gen_Output_File_Ada;

   -----------------------
   -- Gen_Output_File_C --
   -----------------------

   procedure Gen_Output_File_C (Filename : String) is

      Bfile : Name_Id;
      --  Name of generated bind file

   begin
      Create_Binder_Output (Filename, 'c', Bfile);

      --  Write argv/argc stuff if main program case

      if Bind_Main_Program then
         WBI ("extern int gnat_argc;");
         WBI ("extern char **gnat_argv;");
         WBI ("extern char **gnat_envp;");
         WBI ("extern int gnat_exit_status;");
      end if;

      Resolve_Binder_Options;

      --  We need adainit and adafinal for both cases

      Gen_Adainit_C;
      Gen_Adafinal_C;

      --  Main is only present for Ada main case

      if Bind_Main_Program then
         Gen_Main_C;
      end if;

      --  Versions and object files needed in both cases

      Gen_Versions_C;
      Gen_Object_Files_Options;

      --  C binder output is complete

      Close_Binder_Output;
   end Gen_Output_File_C;

   ---------------------
   -- Gen_Output_File --
   ---------------------

   procedure Gen_Output_File (Filename : String) is

      function Public_Version return Boolean;
      --  return true if the version number contains a 'p'

      function Public_Version return Boolean is
      begin
         for I in Gnat_Version_String'Range loop
            if Gnat_Version_String (I) = 'p'  then
               return True;
            end if;
         end loop;
         return False;
      end Public_Version;

   begin
      --  Count number of elaboration calls

      for E in Elab_Order.First .. Elab_Order.Last loop
         if Unit.Table (Elab_Order.Table (E)).No_Elab
           or else Unit.Table (Elab_Order.Table (E)).Shared_Passive
         then
            null;
         else
            Num_Elab_Calls := Num_Elab_Calls + 1;
         end if;
      end loop;

      --  Get the time stamp of the former bind for public version warning

      if Public_Version then
         Record_Time_From_Last_Bind;
      end if;

      --  Generate output file in appropriate language

      if Ada_Bind_File then
         Gen_Output_File_Ada (Filename);
      else
         Gen_Output_File_C (Filename);
      end if;

      --  Periodically issue a warning when the public version is used on
      --  big projects

      if Public_Version then
         Public_Version_Warning;
      end if;
   end Gen_Output_File;

   ----------------------------
   -- Public_Version_Warning --
   ----------------------------

   procedure Public_Version_Warning is

      Time : Int := Time_From_Last_Bind;

      --  Constants to help defining periods

      Always : constant := 0;
      Hour   : constant := 60;
      Day    : constant := 24 * Hour;
      Month  : constant := 30 * Day;
      Never  : constant := Integer'Last;

      --  Constants defining when the warning is issued. For programs more than
      --  'Large' Units they will be issued at 'Period_Large' periods and
      --  for smaller programs they will be issued at 'Period_Small' periods.

      Large        : constant := 20;
      Period_Large : constant := Day;
      Period_Small : constant := Never;

      Nb_Unit : Int;
   begin

      --  Compute the number of units that are not GNAT internal files

      Nb_Unit := 0;
      for A in ALIs.First .. ALIs.Last loop
         if not Is_Internal_File_Name (ALIs.Table (A).Sfile) then
            Nb_Unit := Nb_Unit + 1;
         end if;
      end loop;

      --  Do not emit the message if the last message was emitted in the
      --  specified period taking into account the number of units.

      if Nb_Unit < Large and then Time <= Period_Small then
         return;

      elsif Time <= Period_Large then
         return;
      end if;

      Write_Eol;
      Write_Str ("IMPORTANT NOTICE:");
      Write_Eol;
      Write_Str ("    This version of GNAT is unsupported"
        &                        " and comes with absolutely no warranty.");
      Write_Eol;
      Write_Str ("    If you intend to evaluate or use GNAT for building "
        &                                       "commercial applications,");
      Write_Eol;
      Write_Str ("    please consult http://www.gnat.com/ for information");
      Write_Eol;
      Write_Str ("    on the GNAT Professional product line.");
      Write_Eol;
      Write_Eol;
   end Public_Version_Warning;

   ----------------------
   -- Gen_Versions_Ada --
   ----------------------

   --  This routine generates two sets of lines. The first set has the form:

   --    unnnnn : constant Integer := 16#hhhhhhhh#;

   --  The second set has the form

   --    pragma Export (C, unnnnn, unam);

   --  for each unit, where unam is the unit name suffixed by either B or
   --  S for body or spec, with dots replaced by double underscores, and
   --  hhhhhhhh is the version number, and nnnnn is a 5-digits serial number.

   procedure Gen_Versions_Ada is
      Ubuf : String (1 .. 6) := "u00000";

      procedure Increment_Ubuf;
      --  Little procedure to increment the serial number

      procedure Increment_Ubuf is
      begin
         for J in reverse Ubuf'Range loop
            Ubuf (J) := Character'Succ (Ubuf (J));
            exit when Ubuf (J) <= '9';
            Ubuf (J) := '0';
         end loop;
      end Increment_Ubuf;

   --  Start of processing for Gen_Versions_Ada

   begin
      WBI ("");

      for U in Unit.First .. Unit.Last loop
         Increment_Ubuf;
         WBI ("   " & Ubuf & " : constant Integer := 16#" &
              Unit.Table (U).Version & "#;");
      end loop;

      WBI ("");
      Ubuf := "u00000";

      for U in Unit.First .. Unit.Last loop
         Increment_Ubuf;
         Set_String ("   pragma Export (C, ");
         Set_String (Ubuf);
         Set_String (", """);

         Get_Name_String (Unit.Table (U).Uname);

         for K in 1 .. Name_Len loop
            if Name_Buffer (K) = '.' then
               Set_Char ('_');
               Set_Char ('_');

            elsif Name_Buffer (K) = '%' then
               exit;

            else
               Set_Char (Name_Buffer (K));
            end if;
         end loop;

         if Name_Buffer (Name_Len) = 's' then
            Set_Char ('S');
         else
            Set_Char ('B');
         end if;

         Set_String (""");");
         Write_Statement_Buffer;
      end loop;

   end Gen_Versions_Ada;

   --------------------
   -- Gen_Versions_C --
   --------------------

   --  This routine generates a line of the form:

   --    unsigned unam = 0xhhhhhhhh;

   --  for each unit, where unam is the unit name suffixed by either B or
   --  S for body or spec, with dots replaced by double underscores.

   procedure Gen_Versions_C is
   begin
      for U in Unit.First .. Unit.Last loop
         Set_String ("unsigned ");

         Get_Name_String (Unit.Table (U).Uname);

         for K in 1 .. Name_Len loop
            if Name_Buffer (K) = '.' then
               Set_String ("__");

            elsif Name_Buffer (K) = '%' then
               exit;

            else
               Set_Char (Name_Buffer (K));
            end if;
         end loop;

         if Name_Buffer (Name_Len) = 's' then
            Set_Char ('S');
         else
            Set_Char ('B');
         end if;

         Set_String (" = 0x");
         Set_String (Unit.Table (U).Version);
         Set_Char   (';');
         Write_Statement_Buffer;
      end loop;

   end Gen_Versions_C;

   ----------------------
   -- Lt_Linker_Option --
   ----------------------

   function Lt_Linker_Option (Op1, Op2 : Natural) return Boolean is
   begin
      if Linker_Options.Table (Op1).Internal_File
        /= Linker_Options.Table (Op2).Internal_File
      then
         return Linker_Options.Table (Op1).Internal_File
              < Linker_Options.Table (Op2).Internal_File;
      else
         if Unit.Table (Linker_Options.Table (Op1).Unit).Elab_Position
           /= Unit.Table (Linker_Options.Table (Op2).Unit).Elab_Position
         then
            return Unit.Table (Linker_Options.Table (Op1).Unit).Elab_Position
                 > Unit.Table (Linker_Options.Table (Op2).Unit).Elab_Position;
         else
            return Linker_Options.Table (Op1).Original_Pos
                 < Linker_Options.Table (Op2).Original_Pos;
         end if;
      end if;
   end Lt_Linker_Option;

   ------------------------
   -- Move_Linker_Option --
   ------------------------

   procedure Move_Linker_Option (From : Natural; To : Natural) is
   begin
      Linker_Options.Table (To) := Linker_Options.Table (From);
   end Move_Linker_Option;

   --------------
   -- Set_Char --
   --------------

   procedure Set_Char (C : Character) is
   begin
      Last := Last + 1;
      Statement_Buffer (Last) := C;
   end Set_Char;

   -------------
   -- Set_Int --
   -------------

   procedure Set_Int (N : Int) is
   begin
      if N < 0 then
         Set_String ("-");
         Set_Int (-N);

      else
         if N > 9 then
            Set_Int (N / 10);
         end if;

         Last := Last + 1;
         Statement_Buffer (Last) :=
           Character'Val (N mod 10 + Character'Pos ('0'));
      end if;
   end Set_Int;

   ----------------
   -- Set_String --
   ----------------

   procedure Set_String (S : String) is
   begin
      Statement_Buffer (Last + 1 .. Last + S'Length) := S;
      Last := Last + S'Length;
   end Set_String;

   -------------------
   -- Set_Unit_Name --
   -------------------

   procedure Set_Unit_Name is
   begin
      for J in 1 .. Name_Len - 2 loop
         if Name_Buffer (J) /= '.' then
            Set_Char (Name_Buffer (J));
         else
            Set_String ("__");
         end if;
      end loop;
   end Set_Unit_Name;

   ------------
   -- Tab_To --
   ------------

   procedure Tab_To (N : Natural) is
   begin
      while Last < N loop
         Set_Char (' ');
      end loop;
   end Tab_To;

   ----------------------
   -- Write_Info_Ada_C --
   ----------------------

   procedure Write_Info_Ada_C (Ada : String; C : String; Common : String) is
   begin
      if Ada_Bind_File then
         declare
            S : String (1 .. Ada'Length + Common'Length);

         begin
            S (1 .. Ada'Length) := Ada;
            S (Ada'Length + 1 .. S'Length) := Common;
            WBI (S);
         end;

      else
         declare
            S : String (1 .. C'Length + Common'Length);

         begin
            S (1 .. C'Length) := C;
            S (C'Length + 1 .. S'Length) := Common;
            WBI (S);
         end;
      end if;
   end Write_Info_Ada_C;

   ----------------------------
   -- Write_Statement_Buffer --
   ----------------------------

   procedure Write_Statement_Buffer is
   begin
      WBI (Statement_Buffer (1 .. Last));
      Last := 0;
   end Write_Statement_Buffer;

end Bindgen;
