------------------------------------------------------------------------------
--                                                                          --
--                         GNAT COMPILER COMPONENTS                         --
--                                                                          --
--                             G N A T L I N K                              --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                            $Revision: 1.30 $                             --
--                                                                          --
--          Copyright (C) 1996-1998 Free Software Foundation, Inc.          --
--                                                                          --
-- GNAT is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  GNAT is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNAT;  see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- GNAT was originally developed  by the GNAT team at  New York University. --
-- It is now maintained by Ada Core Technologies Inc (http://www.gnat.com). --
--                                                                          --
------------------------------------------------------------------------------

--  Gnatlink usage: please consult the gnat documentation

with Ada.Command_Line;     use Ada.Command_Line;
with GNAT.OS_Lib;          use GNAT.OS_Lib;
with Gnatvsn;              use Gnatvsn;
with Hostparm;
with Interfaces.C_Streams; use Interfaces.C_Streams;
with Osint;                use Osint;
with Output;               use Output;
with System;               use System;
with Table;

procedure Gnatlink is
   pragma Ident (Gnat_Version_String);

   package Gcc_Linker_Options is new Table.Table (
     Table_Component_Type => String_Access,
     Table_Index_Type     => Integer,
     Table_Low_Bound      => 1,
     Table_Initial        => 20,
     Table_Increment      => 100,
     Table_Name           => "Gnatlink.Gcc_Linker_Options");

   package Linker_Options is new Table.Table (
     Table_Component_Type => String_Access,
     Table_Index_Type     => Integer,
     Table_Low_Bound      => 1,
     Table_Initial        => 20,
     Table_Increment      => 100,
     Table_Name           => "Gnatlink.Linker_Options");

   package Linker_Objects is new Table.Table (
     Table_Component_Type => String_Access,
     Table_Index_Type     => Integer,
     Table_Low_Bound      => 1,
     Table_Initial        => 20,
     Table_Increment      => 100,
     Table_Name           => "Gnatlink.Linker_Objects");

   package Binder_Options is new Table.Table (
     Table_Component_Type => String_Access,
     Table_Index_Type     => Integer,
     Table_Low_Bound      => 1, -- equals low bound of Argument_List for Spawn
     Table_Initial        => 20,
     Table_Increment      => 100,
     Table_Name           => "Gnatlink.Binder_Options");

   Gcc : String_Access := Program_Name ("gcc");

   Read_Mode  : constant String := "r" & Ascii.Nul;

   Begin_Info : String := "/* BEGIN Object file/option list";
   End_Info   : String := "   END Object file/option list */";
   --  Note: above lines are modified in Ada mode, see option processing

   Gcc_Path        : String_Access;
   Linker_Path     : String_Access;

   Output_Filename : String_Access;
   Ali_Filename    : String_Access;
   Binder_Src_File : String_Access;
   Binder_Obj_File : String_Access;

   Tname    : Temp_File_Name;
   Tname_FD : File_Descriptor := Invalid_FD;
   --  Temporary file used by linker to pass list of object files on
   --  certain systems with limitations on size of arguments.

   Debug_Flag_Present : Boolean := False;
   No_Gnat_Lib        : Boolean := False;
   Verbose_Mode       : Boolean := False;

   Ada_Bind_File : Boolean := False;
   --  Set to True if bind file is generated in Ada

   Compile_Bind_File : Boolean := True;
   --  Set to False if bind file is not to be compiled

   function Base_Name (File_Name : in String) return String;
   --  Return just the filename part without the extension (if present).

   procedure Delete (Name : in String);
   --  Wrapper to unlink as status is ignored by this application.

   procedure Exit_With_Error (Error : in String);
   --  Output Error and exit program with a fatal condition.

   function Get_Object_File_Option return String_Access;
   --  Returns a pointer to a string representing the native linker option
   --  which allows a list of object files to be passed to the linker while
   --  contained in a file rather than inline when the linker is invoked.

   procedure Process_Args;
   --  Go through all the arguments and build option tables.

   procedure Process_Binder_File (Name : in String);
   --  Reads the binder file and extracts linker arguments.

   procedure Write_Usage;
   --  Show user the program options.

   ---------------
   -- Base_Name --
   ---------------

   function Base_Name (File_Name : in String) return String is
      Findex1 : Natural;
      Findex2 : Natural;

   begin
      Findex1 := File_Name'First;

      --  The file might be specified by a full path name. However,
      --  we want the path to be stripped away.  In addition to the
      --  default directory_separator allow the '/' to act as separator
      --  since this is allowed in MS-DOS and OS2 ports.

      for J in reverse File_Name'Range loop
         if File_Name (J) = Directory_Separator
           or else File_Name (J) = '/'
         then
            Findex1 := J + 1;
            exit;
         end if;
      end loop;

      Findex2 := File_Name'Last;
      while Findex2 > Findex1
        and then File_Name (Findex2) /=  '.'
      loop
         Findex2 := Findex2 - 1;
      end loop;

      if Findex2 = Findex1 then
         Findex2 := File_Name'Last + 1;
      end if;

      return File_Name (Findex1 .. Findex2 - 1);

   end Base_Name;

   ------------
   -- Delete --
   ------------

   procedure Delete (Name : in String) is
      Status : int;

   begin
      Status := unlink (Name'Address);
   end Delete;

   ---------------------
   -- Exit_With_Error --
   ---------------------

   procedure Exit_With_Error (Error : in String) is
   begin
      Write_Str (Command_Name);
      Write_Str (": ");
      Write_Str (Error);
      Write_Eol;
      Exit_Program (E_Fatal);
   end Exit_With_Error;

   ----------------------------
   -- Get_Object_File_Option --
   ----------------------------

   function Get_Object_File_Option return String_Access is

      procedure Get_Option_Ptr (Length, Ptr : Address);
      pragma Import (C, Get_Option_Ptr, "get_object_file_option_ptr");

      procedure Strncpy (Astring_Addr, Cstring : Address; N : Integer);
      pragma Import (C, Strncpy, "strncpy");

      Option_Ptr    : Address;
      Option_Length : Integer;
      Result        : String_Access;

   begin

      Get_Option_Ptr (Option_Length'Address, Option_Ptr'Address);

      Result := new String (1 .. Option_Length);

      if Option_Length > 0 then
         Strncpy (Result.all'Address, Option_Ptr, Option_Length);
      end if;

      return Result;
   end Get_Object_File_Option;

   ------------------
   -- Process_Args --
   ------------------

   procedure Process_Args is
      Next_Arg : Integer;

   begin
      Binder_Options.Increment_Last;
      Binder_Options.Table (Binder_Options.Last) := new String'("-c");

      Next_Arg := 1;

      loop
         exit when Next_Arg > Argument_Count;

         Process_One_Arg : declare
            Arg : String := Argument (Next_Arg);

         begin
            if Arg'Length /= 0
              and then (Arg (1) = Switch_Character or else Arg (1) = '-')
            then
               if Arg (2) = 'g'
                 and then (Arg'Length = 2
                           or else Arg (3) in '0' .. '3'
                           or else Arg (3 .. Arg'Last) = "coff")
               then
                  Debug_Flag_Present := True;

                  Linker_Options.Increment_Last;
                  Linker_Options.Table (Linker_Options.Last) :=
                   new String'(Arg);

                  Binder_Options.Increment_Last;
                  Binder_Options.Table (Binder_Options.Last) :=
                    Linker_Options.Table (Linker_Options.Last);

               elsif Arg'Length = 2 then
                  case Arg (2) is
                     when 'A' =>
                        Ada_Bind_File := True;
                        Begin_Info := "-- BEGIN Object file/option list";
                        End_Info   := "-- END Object file/option list   ";

                     when 'b' =>
                        Linker_Options.Increment_Last;
                        Linker_Options.Table (Linker_Options.Last) :=
                          new String'(Arg);

                        Binder_Options.Increment_Last;
                        Binder_Options.Table (Binder_Options.Last) :=
                          Linker_Options.Table (Linker_Options.Last);

                        Next_Arg := Next_Arg + 1;

                        if Next_Arg > Argument_Count then
                           Exit_With_Error ("Missing argument for -b");
                        end if;

                        Get_Machine_Name : declare
                           Name_Arg : String_Access :=
                                        new String'(Argument (Next_Arg));

                        begin
                           Linker_Options.Increment_Last;
                           Linker_Options.Table (Linker_Options.Last) :=
                             Name_Arg;

                           Binder_Options.Increment_Last;
                           Binder_Options.Table (Binder_Options.Last) :=
                             Name_Arg;

                        end Get_Machine_Name;

                     when 'C' =>
                        Ada_Bind_File := False;

                     when 'n' =>
                        Compile_Bind_File := False;

                     when 'o' =>
                        Linker_Options.Increment_Last;
                        Linker_Options.Table (Linker_Options.Last) :=
                         new String'(Arg);

                        Next_Arg := Next_Arg + 1;

                        if Next_Arg > Argument_Count then
                           Exit_With_Error ("Missing argument for -o");
                        end if;

                        Output_Filename := new String'(Argument (Next_Arg));

                        Linker_Options.Increment_Last;
                        Linker_Options.Table (Linker_Options.Last) :=
                          Output_Filename;

                     when 'v' =>

                        --  Support "double" verbose mode.  Second -v
                        --  gets sent to the linker and binder phases.

                        if Verbose_Mode then
                           Linker_Options.Increment_Last;
                           Linker_Options.Table (Linker_Options.Last) :=
                            new String'(Arg);

                           Binder_Options.Increment_Last;
                           Binder_Options.Table (Binder_Options.Last) :=
                             Linker_Options.Table (Linker_Options.Last);

                        else
                           Verbose_Mode := True;

                        end if;

                     when others =>
                        Linker_Options.Increment_Last;
                        Linker_Options.Table (Linker_Options.Last) :=
                         new String'(Arg);

                  end case;

               elsif Arg (2) = 'B' then
                  Linker_Options.Increment_Last;
                  Linker_Options.Table (Linker_Options.Last) :=
                   new String'(Arg);

                  Binder_Options.Increment_Last;
                  Binder_Options.Table (Binder_Options.Last) :=
                    Linker_Options.Table (Linker_Options.Last);

               elsif Arg = "-gnatlink" then

                  Next_Arg := Next_Arg + 1;

                  if Next_Arg > Argument_Count then
                     Exit_With_Error ("Missing argument for -gnatlink");
                  end if;

                  Linker_Path :=
                    GNAT.OS_Lib.Locate_Exec_On_Path (Argument (Next_Arg));

                  if Linker_Path = null then
                     Exit_With_Error
                       ("Could not locate linker: " & Argument (Next_Arg));
                  end if;

               elsif Arg = "-nognatlib" then

                  No_Gnat_Lib := True;

               elsif Arg'Length > 6 and then Arg (1 .. 6) = "--GCC=" then
                  declare
                     Program_Args : Argument_List_Access
                       := Argument_String_To_List (Arg (7 .. Arg'Last));

                  begin
                     Gcc := new String'(Program_Args.all (1).all);

                     --  Set appropriate flags for switches passed

                     for I in 2 .. Program_Args.all'Last loop
                        declare
                           Arg : String := Program_Args.all (I).all;
                           AF  : Integer := Arg'First;
                        begin
                           if Arg'Length /= 0
                             and then (Arg (AF) = Switch_Character
                                       or else Arg (AF) = '-')
                           then
                              if Arg (AF + 1) = 'g'
                                and then (Arg'Length = 2
                                  or else Arg (AF + 2) in '0' .. '3'
                                  or else Arg (AF + 2 .. Arg'Last) = "coff")
                              then
                                 Debug_Flag_Present := True;
                              end if;
                           end if;

                           --  Pass to gcc for compiling binder generated file
                           --  No use passing libraries, it will just generate
                           --  a warning

                           if not (Arg (AF .. AF + 1) = "-l"
                             or else Arg (AF .. AF + 1) = "-L")
                           then
                              Binder_Options.Increment_Last;
                              Binder_Options.Table (Binder_Options.Last)
                                := new String'(Arg);
                           end if;

                           --  Pass to gcc for linking program.

                           Gcc_Linker_Options.Increment_Last;
                           Gcc_Linker_Options.Table (Gcc_Linker_Options.Last)
                             := new String'(Arg);
                        end;
                     end loop;

                  end;

               else
                  --  Send all multi-character switches not recognized as
                  --  a special case by gnatlink to the linker/loader stage.

                  Linker_Options.Increment_Last;
                  Linker_Options.Table (Linker_Options.Last) :=
                    new String'(Arg);
               end if;

            else
               if Arg'Length > 4
                 and then Arg (Arg'Last - 3 .. Arg'Last) = ".ali"
               then
                  if Ali_Filename = null then
                     Ali_Filename := new String'(Arg);
                  else
                     Exit_With_Error
                       ("Sorry - cannot handle more than one ALI file");
                  end if;

               elsif Arg'Length > Get_Object_Suffix.all'Length
                 and then Arg
                   (Arg'Last - Get_Object_Suffix.all'Length + 1 .. Arg'Last)
                                                = Get_Object_Suffix.all
               then
                  Linker_Objects.Increment_Last;
                  Linker_Objects.Table (Linker_Objects.Last) :=
                    new String'(Arg);

               else
                  Linker_Options.Increment_Last;
                  Linker_Options.Table (Linker_Options.Last) :=
                    new String'(Arg);
               end if;

            end if;

         end Process_One_Arg;

         Next_Arg := Next_Arg + 1;
      end loop;

      --  If Ada bind file, then compile it with warnings suppressed, because
      --  otherwise the with of the main program may cause junk warnings.

      if Ada_Bind_File then
         Binder_Options.Increment_Last;
         Binder_Options.Table (Binder_Options.Last) := new String'("-gnatws");
      end if;
   end Process_Args;

   -------------------------
   -- Process_Binder_File --
   -------------------------

   procedure Process_Binder_File (Name : in String) is
      Fd          : FILEs;
      Link_Bytes  : Integer := 0;
      Link_Max    : Integer;
      pragma Import (C, Link_Max, "link_max");

      Next_Line   : String (1 .. 1000);
      Nlast       : Integer;
      Objs_Begin  : Integer := 0;
      Objs_End    : Integer := 0;

      Status      : int;
      N           : Integer;

      Object_List_File_Supported : Boolean;
      pragma Import (C, Object_List_File_Supported, "objlist_file_supported");
      --  Predicate indicating whether the linker has an option whereby the
      --  names of object files can be passed to the linker in a file.

      Get_Object_File_Option_Ptr : String_Access;

      procedure Get_Next_Line;
      --  Read the next line from the binder file without the line
      --  terminator.

      procedure Get_Next_Line is
         Fchars : chars;

      begin
         Fchars := fgets (Next_Line'Address, Next_Line'Length, Fd);

         if Fchars = System.Null_Address then
            Exit_With_Error ("Error reading binder output");
         end if;

         Nlast := Next_Line'First;
         while Nlast <= Next_Line'Last
           and then Next_Line (Nlast) /= Ascii.Lf
           and then Next_Line (Nlast) /= Ascii.Cr
         loop
            Nlast := Nlast + 1;
         end loop;

         Nlast := Nlast - 1;
      end Get_Next_Line;

   --  Start of processing for Process_Binder_File

   begin
      Fd := fopen (Name'Address, Read_Mode'Address);

      if Fd = NULL_Stream then
         Exit_With_Error ("Failed to open binder output");
      end if;

      --  Skip up to the Begin Info line

      loop
         Get_Next_Line;
         exit when Next_Line (Next_Line'First .. Nlast) = Begin_Info;
      end loop;

      loop
         Get_Next_Line;
         if Ada_Bind_File then
            Next_Line (Next_Line'First .. Nlast - 8) :=
              Next_Line (Next_Line'First + 8 .. Nlast);
            Nlast := Nlast - 8;
         end if;

         --  Go to next section when switches are reached, the first line
         --  will always start with a "-L"

         exit when Next_Line (1) = '-';

         Linker_Objects.Increment_Last;

         --  Mark the positions of first and last object files in case
         --  they need to be placed with a named file on systems having
         --  linker line limitations.

         if Objs_Begin = 0 then
            Objs_Begin := Linker_Objects.Last;
         end if;

         Linker_Objects.Table (Linker_Objects.Last) :=
           new String'(Next_Line (Next_Line'First .. Nlast));

         Link_Bytes := Link_Bytes + Nlast - Next_Line'First;
      end loop;

      Objs_End := Linker_Objects.Last;

      --  On systems that have limitations on handling very long linker lines
      --  we make use of the system linker option which takes a list of object
      --  file names from a file instead of the command line itself. Our
      --  approach is to replace the list of object files by the special linker
      --  option which then reads the object file list from a file instead.
      --  This method is only triggered if a conservative threshold is passed.

      if Object_List_File_Supported and then Link_Bytes > Link_Max then

         --  Create a temporary file containing the Ada user object files
         --  needed by the link. This list is taken from the bind file
         --  and is output one object per line for maximal compatibility with
         --  linkers supporting this option.

         Create_Temp_File (Tname_FD, Tname);

         for J in Objs_Begin .. Objs_End loop
            Status := Write (Tname_FD, Linker_Objects.Table (J).all'Address,
              Linker_Objects.Table (J).all'Length);
            Status := Write (Tname_FD, Ascii.Lf'Address, 1);
         end loop;

         Close (Tname_FD);

         --  Add the special objects list file option together with the name
         --  of the temporary file to the objects file table.

         Get_Object_File_Option_Ptr := Get_Object_File_Option;
         Linker_Objects.Table (Objs_Begin) :=
           new String'(Get_Object_File_Option_Ptr.all & Tname);

         --  The slots containing these object file names are then removed
         --  from the objects table so they do not appear in the link. They
         --  are removed by moving up the linker options and non-Ada object
         --  files appearing after the Ada object list in the table.

         N := Objs_End - Objs_Begin + 1;
         for J in Objs_End + 1 .. Linker_Objects.Last loop
            Linker_Objects.Table (J - N + 1) := Linker_Objects.Table (J);
         end loop;

         Linker_Objects.Set_Last (Linker_Objects.Last - N + 1);
      end if;

      --  Process switches and options

      loop
         if Next_Line (Next_Line'First .. Nlast) = "-lgnat" then
            if No_Gnat_Lib then
               goto Get_Next_Line_Label;
            end if;

            --  On VMS, add the DECgnat library just before the gnat library.

            if Hostparm.OpenVMS then
               Linker_Options.Increment_Last;

               Linker_Options.Table (Linker_Options.Last) :=
                 new String'("-ldecgnat");
            end if;
         end if;

         Linker_Options.Increment_Last;

         Linker_Options.Table (Linker_Options.Last) :=
           new String'(Next_Line (Next_Line'First .. Nlast));

         <<Get_Next_Line_Label>>
         Get_Next_Line;
         exit when Next_Line (Next_Line'First .. Nlast) = End_Info;

         if Ada_Bind_File then
            Next_Line (Next_Line'First .. Nlast - 8) :=
              Next_Line (Next_Line'First + 8 .. Nlast);
            Nlast := Nlast - 8;
         end if;
      end loop;

      Status := fclose (Fd);

   end Process_Binder_File;

   -----------------
   -- Write_Usage --
   -----------------

   procedure Write_Usage is
   begin
      Write_Str ("Usage: ");
      Write_Str (Command_Name);
      Write_Str (" 'name'.ali");
      Write_Eol;
      Write_Str ("      [-A]               -- binder program in Ada");
      Write_Eol;
      Write_Str ("      [-C]               -- binder program in C (default)");
      Write_Eol;
      Write_Str ("      [-o exec_name]     -- by default it is 'name'");
      Write_Eol;
      Write_Str ("      [-n]               -- do not compile bind file");
      Write_Eol;
      Write_Str ("      [-v]               -- verbose mode");
      Write_Eol;
      Write_Str ("      [-nognatlib]       -- don't add default -lgnat");
      Write_Eol;
      Write_Str ("      [-gnatlink name]   -- full name for the linker (gcc)");
      Write_Eol;
      Write_Str ("      [list of objects]  -- non Ada binaries");
      Write_Eol;
      Write_Str ("      [linker options]   -- other options for the linker");
      Write_Eol;
   end Write_Usage;

--  Start of processing for Gnatlink

begin
   if Argument_Count = 0 then
      Write_Usage;
      Exit_Program (E_Fatal);
   end if;

   Process_Args;

   --  Locate all the necessary programs and verify required files are present

   Gcc_Path := GNAT.OS_Lib.Locate_Exec_On_Path (Gcc.all);

   if Gcc_Path = null then
      Exit_With_Error ("Couldn't locate " & Gcc.all);
   end if;

   if Linker_Path = null then
      Linker_Path := Gcc_Path;
   end if;

   if Ali_Filename = null then
      Exit_With_Error ("Required 'name'.ali not present.");
   end if;

   if not Is_Regular_File (Ali_Filename.all) then
      Exit_With_Error (Ali_Filename.all & " not found.");
   end if;

   if Verbose_Mode then
      Write_Eol;
      Write_Str ("GNATLINK ");
      Write_Str (Gnat_Version_String);
      Write_Str (" Copyright 1996-1998 Free Software Foundation, Inc.");
      Write_Eol;
   end if;

   --  If there wasn't an output specified, then use the base name of
   --  the .ali filename.

   if Output_Filename = null then

      Output_Filename :=
        new String'(Base_Name (Ali_Filename.all)
                       & Get_Debuggable_Suffix.all);

      Linker_Options.Increment_Last;
      Linker_Options.Table (Linker_Options.Last) :=
        new String'("-o");

      Linker_Options.Increment_Last;
      Linker_Options.Table (Linker_Options.Last) :=
        new String'(Output_Filename.all);

   end if;

   --  Transform the .ali filename into the binder output filename.

   Make_Binder_Filenames : declare
      Fname     : String  := Base_Name (Ali_Filename.all);
      Fname_Len : Integer := Fname'Length;

      function Get_Maximum_File_Name_Length return Integer;
      pragma Import (C, Get_Maximum_File_Name_Length,
                        "Get_Maximum_File_Name_Length");

      Maximum_File_Name_Length : Integer := Get_Maximum_File_Name_Length;

      Second_Char : Character;
      --  Second character of name of files

   begin
      --  Set proper second character of file name

      if not Ada_Bind_File then
         Second_Char := '_';

      elsif Hostparm.OpenVMS then
         Second_Char := '$';

      else
         Second_Char := '~';
      end if;

      --  If the length of the binder file becomes too long due to
      --  the addition of the "b?" prefix, then truncate it.

      if Maximum_File_Name_Length > 0 then
         while Fname_Len > Maximum_File_Name_Length - 2 loop
            Fname_Len := Fname_Len - 1;
         end loop;
      end if;

      if Ada_Bind_File then
         Binder_Src_File :=
           new String'('b'
                       & Second_Char
                       & Fname (Fname'First .. Fname'First + Fname_Len - 1)
                       & ".adb");

      else
         Binder_Src_File :=
           new String'('b'
                       & Second_Char
                       & Fname (Fname'First .. Fname'First + Fname_Len - 1)
                       & ".c");
      end if;

      Binder_Obj_File :=
        new String'('b'
                    & Second_Char
                    & Fname (Fname'First .. Fname'First + Fname_Len - 1)
                    & Get_Object_Suffix.all);

      if Fname_Len /= Fname'Length then
         Binder_Options.Increment_Last;
         Binder_Options.Table (Binder_Options.Last) := new String'("-o");
         Binder_Options.Increment_Last;
         Binder_Options.Table (Binder_Options.Last) := Binder_Obj_File;
      end if;

   end Make_Binder_Filenames;

   Process_Binder_File (Binder_Src_File.all & Ascii.Nul);

   --  Compile the binder file. This is fast, so we always do it, unless
   --  specifically told not to by the -n switch

   if Compile_Bind_File then
      Bind_Step : declare
         Success : Boolean;
         Args    : Argument_List (1 .. Binder_Options.Last + 1);

      begin
         for J in Binder_Options.First .. Binder_Options.Last loop
            Args (J) := Binder_Options.Table (J);
         end loop;

         Args (Args'Last) := Binder_Src_File;

         if Verbose_Mode then
            Write_Str (Base_Name (Gcc_Path.all));

            for J in Args'Range loop
               Write_Str (" ");
               Write_Str (Args (J).all);
            end loop;

            Write_Eol;
         end if;

         GNAT.OS_Lib.Spawn (Gcc_Path.all, Args, Success);

         if not Success then
            Exit_Program (E_Fatal);
         end if;
      end Bind_Step;
   end if;

   --  Now, actually link the program.

   Link_Step : declare
      Success  : Boolean;
      Num_Args : constant Integer :=
                  (Linker_Options.Last - Linker_Options.First + 1) +
                  (Gcc_Linker_Options.Last - Gcc_Linker_Options.First + 1) +
                  (Linker_Objects.Last - Linker_Objects.First + 1);
      Args     : Argument_List (1 .. Num_Args + 1);
      Index    : Integer := Args'First;

   begin
      Args (Index) := Binder_Obj_File;

      --  Add the object files and any -largs libraries

      for J in Linker_Objects.First .. Linker_Objects.Last loop
         Index := Index + 1;
         Args (Index) := Linker_Objects.Table (J);
      end loop;

      --  Add the linker options from the binder file

      for J in Linker_Options.First .. Linker_Options.Last loop
         Index := Index + 1;
         Args (Index) := Linker_Options.Table (J);
      end loop;

      --  Finally add the libraries from the --GCC= switch

      for J in Gcc_Linker_Options.First .. Gcc_Linker_Options.Last loop
         Index := Index + 1;
         Args (Index) := Gcc_Linker_Options.Table (J);
      end loop;

      if Verbose_Mode then
         Write_Str (Base_Name (Linker_Path.all));

         for J in Args'Range loop
            Write_Str (" ");
            Write_Str (Args (J).all);
         end loop;

         Write_Eol;
      end if;

      GNAT.OS_Lib.Spawn (Linker_Path.all, Args, Success);

      --  Delete the temporary file used in conjuction with linking if
      --  one was created. See Process_Bind_File for details.

      if Tname_FD /= Invalid_FD then
         Delete (Tname);
      end if;

      if not Success then
         Exit_Program (E_Fatal);
      end if;

   end Link_Step;

   --  Only keep the binder output file and it's associated object
   --  file if compiling with the -g option.  These files are only
   --  useful if debugging.

   if not Debug_Flag_Present then
      Delete (Binder_Src_File.all & Ascii.Nul);
      Delete (Binder_Obj_File.all & Ascii.Nul);
   end if;

   Exit_Program (E_Success);

exception
   when others =>
      Exit_With_Error ("INTERNAL ERROR. Please report.");
end Gnatlink;
