------------------------------------------------------------------------------
--                                                                          --
--                GNU ADA RUN-TIME LIBRARY (GNARL) COMPONENTS               --
--                                                                          --
--                     S Y S T E M . T A S K _ T I M E R                    --
--                                                                          --
--                                  B o d y                                 --
--                                                                          --
--                             $Revision: 1.31 $                            --
--                                                                          --
--            Copyright (C) 1991-1998, Florida State University             --
--                                                                          --
-- GNARL is free software; you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion. GNARL is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNARL; see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable  to  be --
-- covered  by the  GNU  General  Public  License.  This exception does not --
-- however invalidate  any other reasons why  the executable file  might be --
-- covered by the  GNU Public License.                                      --
--                                                                          --
-- GNARL was developed by the GNARL team at Florida State University. It is --
-- now maintained by Ada Core Technologies Inc. in cooperation with Florida --
-- State University (http://www.gnat.com).                                  --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Calendar;
--  Used for Time

with Ada.Calendar.Delays;
--  Used for To_Duration

with Ada.Real_Time;
--  Used for Time
--           Time_Span
--           To_Duration

with Ada.Real_Time.Delays;
--  Used for To_Duration

with System.Tasking;
--  Used for Task_ID
--           Timer_Server_Sleep

with System.Task_Primitives.Operations;
--  Used for Timed_Sleep
--           Write_Lock
--           Unlock
--           Self
--           Clock

with System.Tasking.Utilities;
--  Used for Make_Independent

with System.Tasking.Initialization;
--  Used for Defer_Abort
--           Undefer_Abort

with System.OS_Primitives;
--  used for Max_Sensible_Delay
--           Delay_Modes

package body System.Task_Timer is

   package STPO renames System.Task_Primitives.Operations;
   package ST renames System.Tasking;
   package STU renames System.Tasking.Utilities;
   package STI renames System.Tasking.Initialization;
   package OSP renames System.OS_Primitives;

   Max_Clock_Val : constant Duration := STPO.Clock + OSP.Max_Sensible_Delay;
   --  This constant represent the infinite time so that delaying
   --  until this time means sleeping forever (We do not expect
   --  one to use a delay of more than Max_Sensible_Delay).

   Timer_Server_ID : ST.Task_ID;

   Timer_Attention : Boolean := False;
   pragma Atomic (Timer_Attention);

   -------------------
   -- Signal_Object --
   -------------------

   protected body Signal_Object is

      entry Wait when Open is
      begin
         Open := False;
      end Wait;

      procedure Signal is
      begin
         Open := True;
      end Signal;

   end Signal_Object;

   task type Timer_Server_Type is
      pragma Interrupt_Priority (System.Any_Priority'Last);
   end Timer_Server_Type;

   type Timer_Server_Type_Ptr is access Timer_Server_Type;

   Timer_Server : Timer_Server_Type_Ptr;

   Q_Head : Q_Link := null;

   -----------
   -- Timer --
   -----------

   protected body Timer is

      ------------------------
      -- Timer.Time_Enqueue --
      ------------------------

      --  Allocate a queue element for the wakeup time T and put it in the
      --  queue in wakeup time order. Return the allocated queue element
      --  in N.

      procedure Time_Enqueue
        (T : Duration;
         D : access Delay_Block)
      is
         Q_Ptr : Q_Link := Q_Head;
         N     : constant Q_Link := Q_Link (D);

         use type ST.Task_ID;
         --  for visibility of operator "="

      begin
         if Timer_Server = null then
            Timer_Server := new Timer_Server_Type;

            --  ?????
            --  The above dynamic creation of a timer server task is intended
            --  to avoid the overhead of creating this task for programs that
            --  do not use select-then-abort with time delays. However, there
            --  are some other reasons why it is a BAD idea. First of all, the
            --  time to create and activate the server may be longer than the
            --  duration requested by the user. Second, this creates an
            --  anomalous situation that will defeat important internal
            --  consistency checks of the runtime system.
            --  Normally, task creation occurs from user-level code, without
            --  abort deferral. There are Asserts throughout the run-time
            --  system to verify this, since task creation is a blocking
            --  operation (the creator blocks to await activation of the child)
            --  that we do not want users to peform within normally
            --  abort-deferred regions (e.g., protected operations and
            --  finalizers).

            while Timer_Server_ID = null loop
               STPO.Yield;
            end loop;
         end if;

         N.T := T;

         --  If the new element becomes head of the queue, notify Timer Service

         if Q_Head = null then
            N.Next := null;
            N.Previous := null;
            Q_Head := N;
            --  Signal the timer server to wake up
            STPO.Write_Lock (Timer_Server_ID);
            Timer_Attention := True;
            STPO.Wakeup
              (Timer_Server_ID, ST.Timer_Server_Sleep);
            STPO.Unlock (Timer_Server_ID);

         elsif N.T < Q_Head.T then
            N.Next := Q_Head;
            N.Previous := null;
            Q_Head.Previous := N;
            Q_Head := N;
            --  Signal the timer server to wake up
            STPO.Write_Lock (Timer_Server_ID);
            Timer_Attention := True;
            STPO.Wakeup (Timer_Server_ID, ST.Timer_Server_Sleep);
            STPO.Unlock (Timer_Server_ID);

         else
            --  Place in the middle

            while Q_Ptr.Next /= null loop
               if Q_Ptr.Next.T >= N.T then
                  N.Next := Q_Ptr.Next;
                  N.Previous := Q_Ptr;
                  Q_Ptr.Next.Previous := N;
                  Q_Ptr.Next := N;
                  exit;
               end if;
               Q_Ptr := Q_Ptr.Next;
            end loop;

            if Q_Ptr.Next = null then
               --  Place at the end
               N.Next := null;
               N.Previous := Q_Ptr;
               Q_Ptr.Next := N;
            end if;
         end if;
      end Time_Enqueue;

      -------------------
      -- Timer.Service --
      -------------------

      --  Service all of the wakeup requests on the queue whose wakeup time
      --  have reached. Return the next wakeup time
      --  after that (the wakeup time of the head of the queue if any;
      --  a time far in the future if not).

      procedure Service (T : out Duration) is
         Q_Ptr : Q_Link := Q_Head;
         Now   : constant Duration := STPO.Clock;

      begin
         Timer_Attention := False;
         while Q_Ptr /= null and then Q_Ptr.T <= Now loop
            --  Wake up the waiting task
            pragma Debug (System.Tasking.Debug.Trace
              ("Service: waking up waiting task", 'E'));
            Q_Ptr.S_O.Signal;
            Dequeue (Q_Ptr);
            --  Remove the entry
            Q_Ptr := Q_Ptr.Next;
         end loop;
         if Q_Head = null then
            T := Max_Clock_Val;
         else
            T := Q_Head.T;
         end if;
      end Service;

      -------------------
      -- Timer.Dequeue --
      -------------------

      procedure Dequeue (D : access Delay_Block) is
         Q_Ptr : constant Q_Link := Q_Link (D);

      begin
         --  Case of head entry
         if Q_Head = Q_Ptr then
            Q_Head := Q_Ptr.Next;
            if Q_Head /= null then
               Q_Head.Previous := null;
            end if;
         --  Case of tail entry
         elsif Q_Ptr.Next = null then
            if Q_Ptr.Previous /= null then
               Q_Ptr.Previous.Next := null;
            end if;
         else
            Q_Ptr.Previous.Next := Q_Ptr.Next;
            Q_Ptr.Next.Previous := Q_Ptr.Previous;
         end if;
         Q_Ptr.Next := null;
         Q_Ptr.Previous := null;
      end Dequeue;

      -----------------
      -- Timer.Empty --
      -----------------

      function Empty return Boolean is
      begin
         return Q_Head = null;
      end Empty;


      --  ??? The following entries used to all be called Enqueue; the
      --      compiler does not seem to be able to handle overloading
      --      in requeue statements.

      -----------------------------
      -- Timer.Enqueue_Time_Span --
      -----------------------------

      entry Enqueue_Time_Span
        (T : in Ada.Real_Time.Time_Span;
         D : access Delay_Block)
      when True is
         N : constant Q_Link := Q_Link (D);
      begin
         Time_Enqueue (STPO.Clock + Duration'Min
           (Ada.Real_Time.To_Duration (T), OSP.Max_Sensible_Delay), D);
         requeue N.S_O.Wait with abort;
      end Enqueue_Time_Span;

      entry Enqueue_Duration
         (T : in Duration;
          D : access Delay_Block)
      when True is
         N : constant Q_Link := Q_Link (D);
      begin
         Time_Enqueue
           (STPO.Clock + Duration'Min (T, OSP.Max_Sensible_Delay), D);
         requeue N.S_O.Wait with abort;
      end Enqueue_Duration;

      entry Enqueue_Real_Time
        (T : in Ada.Real_Time.Time;
         D : access Delay_Block)
      when True is
         N : constant Q_Link := Q_Link (D);
      begin
         Time_Enqueue (Ada.Real_Time.Delays.To_Duration (T), D);
         requeue N.S_O.Wait with abort;
      end Enqueue_Real_Time;

      entry Enqueue_Calendar_Time
        (T : in Ada.Calendar.Time;
         D : access Delay_Block)
      when True is
         N : constant Q_Link := Q_Link (D);
      begin
         Time_Enqueue (Ada.Calendar.Delays.To_Duration (T), D);
         requeue N.S_O.Wait with abort;
      end Enqueue_Calendar_Time;

   end Timer;

   --------------
   -- Finalize --
   --------------

   procedure Finalize (Object : in out Delay_Block) is
   begin
      Timer.Dequeue (Object'Access);
   end Finalize;

   ------------------
   -- Timer_Server --
   ------------------

   task body Timer_Server_Type is
      Next_Wakeup_Time : Duration := Max_Clock_Val;
      Timedout         : Boolean;
      Yielded          : Boolean;

   begin
      Timer_Server_ID := STPO.Self;

      STU.Make_Independent;

      loop
         --  The timer server needs to catch pending aborts after
         --  finalization of library packages.
         --  If it doesn't poll for it, the server will
         --  sometimes hang.

         STI.Defer_Abort (Timer_Server_ID);
         STPO.Write_Lock (Timer_Server_ID);

         if not Timer_Attention then
            Timer_Server_ID.State := ST.Timer_Server_Sleep;
            if Next_Wakeup_Time = Max_Clock_Val then
               Timer_Server_ID.User_State := 1;
            else
               Timer_Server_ID.User_State := 2;
            end if;

            STPO.Timed_Sleep
              (Timer_Server_ID, Next_Wakeup_Time,
               OSP.Absolute_RT, ST.Timer_Server_Sleep,
               Timedout, Yielded);
            Timer_Server_ID.State := ST.Runnable;
         end if;

         STPO.Unlock (Timer_Server_ID);

         if not Yielded then
            STPO.Yield;
         end if;

         STI.Undefer_Abort (Timer_Server_ID);

         Timer_Server_ID.User_State := 3;
         Timer.Service (Next_Wakeup_Time);

         --  Timer.Service returns the Next_Wakeup_Time.
         --  The Next_Wakeup_Time is either an infinity (no delay request)
         --  or the wakeup time of the queue head. This value is used for
         --  an actual delay in this server.

      end loop;
   end Timer_Server_Type;

end System.Task_Timer;
