------------------------------------------------------------------------------
--                                                                          --
--                         GNAT COMPILER COMPONENTS                         --
--                                                                          --
--                             X R  _ T A B L S                             --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                            $Revision: 1.14 $                             --
--                                                                          --
--          Copyright (C) 1998 Free Software Foundation, Inc.               --
--                                                                          --
-- GNAT is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  GNAT is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNAT;  see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- GNAT is maintained by Ada Core Technologies Inc (http://www.gnat.com).   --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Strings.Fixed;
with Ada.Strings;
with Ada.Text_IO;
with Ada.Characters.Handling;
with Interfaces.C_Streams;
with Interfaces.C;
with System;
with Hostparm;
with GNAT.IO_Aux;
with Unchecked_Deallocation;

package body Xr_Tabls is

   subtype Line_String      is String (1 .. Hostparm.Max_Line_Length);
   subtype File_Name_String is String (1 .. Hostparm.Max_Name_Length);

   Files         : File_Table;
   Entities      : Entity_Table;
   Directories   : Project_File_Ptr;
   Default_Match : Boolean := False;

   ---------------------
   -- Add_Declaration --
   ---------------------

   function Add_Declaration
     (File_Ref  : File_Reference;
      Symbol    : String;
      Line      : Natural;
      Column    : Natural;
      Decl_Type : Character)
      return      Declaration_Reference
   is
      The_Entities : Declaration_Reference := Entities.Table;
      New_Decl     : Declaration_Reference;
      Result       : Compare_Result;
      Prev         : Declaration_Reference := null;

   begin
      --  Check if the identifier already exists in the table

      while The_Entities /= null loop
         Result := Compare (The_Entities, File_Ref, Line, Column, Symbol);
         exit when Result = GreaterThan;

         if Result = Equal then
            return The_Entities;
         end if;

         Prev := The_Entities;
         The_Entities  := The_Entities.Next;
      end loop;

      --  Insert the Declaration in the table

      New_Decl := new Declaration_Record'
        (Symbol_Length => Symbol'Length,
         Symbol        => Symbol,
         File          => File_Ref,
         Line          => Line,
         Column        => Column,
         Decl_Type     => Decl_Type,
         Body_Ref      => null,
         Ref_Ref       => null,
         Modif_Ref     => null,
         Match         => Default_Match or else Match (File_Ref, Line, Column),
         Next          => null);

      if Prev = null then
         New_Decl.Next := Entities.Table;
         Entities.Table := New_Decl;
      else
         New_Decl.Next := Prev.Next;
         Prev.Next := New_Decl;
      end if;

      return New_Decl;
   end Add_Declaration;

   --------------
   -- Add_File --
   --------------

   procedure Add_File
     (File_Name    : String;
      File_Existed : out Boolean;
      Ref          : out File_Reference;
      Visited      : Boolean := True;
      Emit_Warning : Boolean := False)
   is
      The_Files : File_Reference := Files.Table;

   begin
      --  Check if the file already exists in the table

      while The_Files /= null loop

         if The_Files.File = File_Name then
            File_Existed      := True;
            Ref               := The_Files;
            return;
         end if;

         The_Files := The_Files.Next;
      end loop;

      Ref := new File_Record'(File_Length  => File_Name'Length,
                              File         => File_Name,
                              Dir          => null,
                              Lines        => null,
                              Visited      => Visited,
                              Emit_Warning => Emit_Warning,
                              Next         => Files.Table);
      Files.Table := Ref;
      File_Existed := False;
   end Add_File;

   --------------
   -- Add_Line --
   --------------

   procedure Add_Line
     (File   : File_Reference;
      Line   : Natural;
      Column : Natural)
   is
   begin
      File.Lines := new Ref_In_File'(Line   => Line,
                                     Column => Column,
                                     Next   => File.Lines);
   end Add_Line;

   -------------------
   -- Add_Reference --
   -------------------

   procedure Add_Reference
     (Declaration : Declaration_Reference;
      File_Ref    : File_Reference;
      Line        : Natural;
      Column      : Natural;
      Ref_Type    : Character)
   is
      procedure Free is new Unchecked_Deallocation
        (Reference_Record, Reference);

      Ref     : Reference;
      Prev    : Reference := null;
      Result  : Compare_Result;
      New_Ref : Reference := new Reference_Record'(File   => File_Ref,
                                                   Line   => Line,
                                                   Column => Column,
                                                   Next   => null);
   begin
      case Ref_Type is
         when 'b' => Ref := Declaration.Body_Ref;
         when 'r' => Ref := Declaration.Ref_Ref;
         when 'm' => Ref := Declaration.Modif_Ref;
         when others => return;
      end case;

      --  Check if the reference already exists

      while Ref /= null loop
         Result := Compare (New_Ref, Ref);
         exit when Result = LessThan;

         if Result = Equal then
            Free (New_Ref);
            return;
         end if;

         Prev := Ref;
         Ref  := Ref.Next;
      end loop;

      --  Insert it in the list

      if Prev /= null then
         New_Ref.Next := Prev.Next;
         Prev.Next := New_Ref;

      else
         case Ref_Type is
            when 'b' =>
               New_Ref.Next := Declaration.Body_Ref;
               Declaration.Body_Ref := New_Ref;
            when 'r' =>
               New_Ref.Next := Declaration.Ref_Ref;
               Declaration.Ref_Ref := New_Ref;
            when 'm' =>
               New_Ref.Next := Declaration.Modif_Ref;
               Declaration.Modif_Ref := New_Ref;
            when others => null;
         end case;
      end if;

      if not Declaration.Match then
         Declaration.Match := Match (File_Ref, Line, Column);
      end if;
   end Add_Reference;

   -------------------
   -- ALI_File_Name --
   -------------------

   function ALI_File_Name (Ada_File_Name : String) return String is
      Index : Natural := Ada.Strings.Fixed.Index
        (Ada_File_Name, ".", Going => Ada.Strings.Backward);
   begin
      if Index /= 0 then
         return Ada_File_Name (Ada_File_Name'First .. Index)
           & "ali";
      else
         return Ada_File_Name & ".ali";
      end if;
   end ALI_File_Name;

   -------------
   -- Compare --
   -------------

   function Compare
     (Ref1 : Reference;
      Ref2 : Reference)
      return Compare_Result
   is
   begin
      if Ref1 = null then
         return GreaterThan;
      elsif Ref2 = null then
         return LessThan;
      end if;

      if Ref1.File.File < Ref2.File.File then
         return LessThan;

      elsif Ref1.File.File = Ref2.File.File then
         if Ref1.Line < Ref2.Line then
            return LessThan;

         elsif Ref1.Line = Ref2.Line then
            if Ref1.Column < Ref2.Column then
               return LessThan;
            elsif Ref1.Column = Ref2.Column then
               return Equal;
            else
               return GreaterThan;
            end if;

         else
            return GreaterThan;
         end if;

      else
         return GreaterThan;
      end if;
   end Compare;

   -------------
   -- Compare --
   -------------

   function Compare
     (Decl1 : Declaration_Reference;
      File2 : File_Reference;
      Line2 : Integer;
      Col2  : Integer;
      Symb2 : String)
      return  Compare_Result
   is
   begin
      if Decl1 = null then
         return GreaterThan;
      end if;

      if Decl1.File.File < Get_File (File2) then
         return LessThan;

      elsif Decl1.File.File = Get_File (File2) then
         if Decl1.Line < Line2 then
            return LessThan;

         elsif Decl1.Line = Line2 then
            if Decl1.Column < Col2 then
               return LessThan;

            elsif Decl1.Column = Col2 then
               if Decl1.Symbol < Symb2 then
                  return LessThan;
               elsif Decl1.Symbol > Symb2 then
                  return GreaterThan;
               end if;

               return Equal;

            else
               return GreaterThan;
            end if;

         else
            return GreaterThan;
         end if;

      else
         return GreaterThan;
      end if;
   end Compare;

   -------------------------
   -- Create_Project_File --
   -------------------------

   procedure Create_Project_File
     (Name           : String;
      Include_Dir    : String)
   is
      use Ada.Strings.Unbounded;

      Line        : File_Name_String;
      Line_Length : Natural;
      Obj_Dir     : Unbounded_String := Null_Unbounded_String;
      Src_Dir     : Unbounded_String := Null_Unbounded_String;
      File        : Ada.Text_IO.File_Type;

      Gnatls_Src_Cache : Unbounded_String;
      Gnatls_Obj_Cache : Unbounded_String;

   begin
      --  If we have a project file name then parse the file

      if Name'Length /= 0 then
         Ada.Text_IO.Open (File => File,
                           Mode => Ada.Text_IO.In_File,
                           Name => Name);

         while not Ada.Text_IO.End_Of_File (File) loop
            Ada.Text_IO.Get_Line (File, Line, Line_Length);

            if Line_Length >= Line'First + 7 then
               declare
                  S : String := Ada.Strings.Fixed.Trim
                                  (Line (9 .. Line_Length), Ada.Strings.Both);

               begin
                  if Line (1 .. 8) = "obj_dir=" then
                     if S (S'Last) = '/' then
                        Append (Obj_Dir, S & " ");
                     else
                        Append (Obj_Dir, S & "/ ");
                     end if;

                  elsif Line (1 .. 8) = "src_dir=" then
                     if S (S'Last) = '/' then
                        Append (Src_Dir, S & " ");
                     else
                        Append (Src_Dir, S & "/ ");
                     end if;
                  end if;
               end;
            end if;
         end loop;
      end if;

      if Ada.Text_IO.Is_Open (File) then
         Ada.Text_IO.Close (File);
      end if;

      Parse_Gnatls (Include_Dir, Gnatls_Src_Cache, Gnatls_Obj_Cache);

      Directories := new Project_File'
        (Src_Dir_Length     => Length (Src_Dir) + Length (Gnatls_Src_Cache),
         Obj_Dir_Length     => Length (Obj_Dir) + Length (Gnatls_Obj_Cache),
         Src_Dir            => To_String (Src_Dir & Gnatls_Src_Cache),
         Obj_Dir            => To_String (Obj_Dir & Gnatls_Obj_Cache),
         Src_Dir_Index      => 1,
         Obj_Dir_Index      => 1,
         Last_Obj_Dir_Start => 0);
   end Create_Project_File;

   ---------------------
   -- Current_Obj_Dir --
   ---------------------

   function Current_Obj_Dir return String is
   begin
      return Directories.Obj_Dir (Directories.Last_Obj_Dir_Start
                                  .. Directories.Obj_Dir_Index - 2);
   end Current_Obj_Dir;

   -------------------
   -- Find_ALI_File --
   -------------------

   function Find_ALI_File (Short_Name  : String) return String is
      use type Ada.Strings.Unbounded.String_Access;
      Old_Obj_Dir : constant Integer := Directories.Obj_Dir_Index;
   begin
      Reset_Obj_Dir;

      loop
         declare
            Obj_Dir : String := Next_Obj_Dir;
         begin
            exit when Obj_Dir'Length = 0;
            if GNAT.IO_Aux.File_Exists (Obj_Dir & Short_Name) then
               Directories.Obj_Dir_Index := Old_Obj_Dir;
               return Obj_Dir;
            end if;
         end;
      end loop;

      --  Finally look in the standard directories

      Directories.Obj_Dir_Index := Old_Obj_Dir;
      return "";
   end Find_ALI_File;

   ----------------------
   -- Find_Source_File --
   ----------------------

   function Find_Source_File (Short_Name  : String) return String is
      use type Ada.Strings.Unbounded.String_Access;

   begin
      Reset_Src_Dir;
      loop
         declare
            Src_Dir : String := Next_Src_Dir;
         begin
            exit when Src_Dir'Length = 0;

            if GNAT.IO_Aux.File_Exists (Src_Dir & Short_Name) then
               return Src_Dir;
            end if;
         end;
      end loop;

      --  Finally look in the standard directories

      return "";
   end Find_Source_File;

   ----------------
   -- First_Body --
   ----------------

   function First_Body (Decl : Declaration_Reference) return Reference is
   begin
      return Decl.Body_Ref;
   end First_Body;

   -----------------------
   -- First_Declaration --
   -----------------------

   function First_Declaration return Declaration_Reference is
   begin
      return Entities.Table;
   end First_Declaration;

   -----------------
   -- First_Modif --
   -----------------

   function First_Modif (Decl : Declaration_Reference) return Reference is
   begin
      return Decl.Modif_Ref;
   end First_Modif;

   ---------------------
   -- First_Reference --
   ---------------------

   function First_Reference (Decl : Declaration_Reference) return Reference is
   begin
      return Decl.Ref_Ref;
   end First_Reference;

   ----------------
   -- Get_Column --
   ----------------

   function Get_Column (Decl : Declaration_Reference) return String is
   begin
      return Ada.Strings.Fixed.Trim (Natural'Image (Decl.Column),
                                     Ada.Strings.Left);
   end Get_Column;

   function Get_Column (Ref : Reference) return String is
   begin
      return Ada.Strings.Fixed.Trim (Natural'Image (Ref.Column),
                                     Ada.Strings.Left);
   end Get_Column;

   ----------------------
   -- Get_Emit_Warning --
   ----------------------

   function Get_Emit_Warning (File : File_Reference) return Boolean is
   begin
      return File.Emit_Warning;
   end Get_Emit_Warning;

   --------------
   -- Get_File --
   --------------

   function Get_File
     (Decl     : Declaration_Reference;
      With_Dir : Boolean := False)
      return     String
   is
   begin
      return Get_File (Decl.File, With_Dir);
   end Get_File;

   function Get_File
     (Ref      : Reference;
      With_Dir : Boolean := False)
      return     String
   is
   begin
      return Get_File (Ref.File, With_Dir);
   end Get_File;

   function Get_File
     (File     : File_Reference;
      With_Dir : in Boolean := False)
      return     String
   is
   begin
      --  If we do not want the full path name

      if not With_Dir then
         return File.File;
      end if;

      if File.Dir = null then

         if Ada.Strings.Fixed.Tail (File.File, 3) = "ali" then
            File.Dir := new String'(Find_ALI_File (File.File));
         else
            File.Dir := new String'(Find_Source_File (File.File));
         end if;
      end if;

      return File.Dir.all & File.File;
   end Get_File;

   ------------------
   -- Get_File_Ref --
   ------------------

   function Get_File_Ref (Ref : Reference) return File_Reference is
   begin
      return Ref.File;
   end Get_File_Ref;

   --------------
   -- Get_Line --
   --------------

   function Get_Line (Decl : Declaration_Reference) return String is
   begin
      return Ada.Strings.Fixed.Trim (Natural'Image (Decl.Line),
                                     Ada.Strings.Left);
   end Get_Line;

   function Get_Line (Ref : Reference) return String is
   begin
      return Ada.Strings.Fixed.Trim (Natural'Image (Ref.Line),
                                     Ada.Strings.Left);
   end Get_Line;

   ----------------
   -- Get_Symbol --
   ----------------

   function Get_Symbol (Decl : Declaration_Reference) return String is
   begin
      return Decl.Symbol;
   end Get_Symbol;

   --------------
   -- Get_Type --
   --------------

   function Get_Type (Decl : Declaration_Reference) return String is
   begin
      return String'(1 => Ada.Characters.Handling.To_Upper (Decl.Decl_Type));
   end Get_Type;

   -----------
   -- Match --
   -----------

   function Match
     (File   : File_Reference;
      Line   : Natural;
      Column : Natural)
      return   Boolean
   is
      Ref : Ref_In_File_Ptr := File.Lines;

   begin
      while Ref /= null loop
         if (Ref.Line = 0 or else Ref.Line = Line)
           and then (Ref.Column = 0 or else Ref.Column = Column)
         then
            return True;
         end if;

         Ref := Ref.Next;
      end loop;

      return False;
   end Match;

   -----------
   -- Match --
   -----------

   function Match (Decl : Declaration_Reference) return Boolean is
   begin
      return Decl.Match;
   end Match;

   ----------
   -- Next --
   ----------

   function Next (Decl : Declaration_Reference) return Declaration_Reference is
   begin
      return Decl.Next;
   end Next;

   ----------
   -- Next --
   ----------

   function Next (Ref : Reference) return Reference is
   begin
      return Ref.Next;
   end Next;

   ------------------
   -- Next_Src_Dir --
   ------------------

   function Next_Src_Dir return String is
      First : Integer := Directories.Src_Dir_Index;
      Last  : Integer := Directories.Src_Dir_Index;

   begin
      if Last > Directories.Src_Dir_Length then
         return String'(1 .. 0 => ' ');
      end if;

      while Directories.Src_Dir (Last) /= ' ' loop
         Last := Last + 1;
      end loop;

      Directories.Src_Dir_Index := Last + 1;
      return Directories.Src_Dir (First .. Last - 1);
   end Next_Src_Dir;

   ------------------
   -- Next_Obj_Dir --
   ------------------

   function Next_Obj_Dir return String is
      First : Integer := Directories.Obj_Dir_Index;
      Last  : Integer := Directories.Obj_Dir_Index;

   begin
      if Last > Directories.Obj_Dir_Length then
         return String'(1 .. 0 => ' ');
      end if;

      while Directories.Obj_Dir (Last) /= ' ' loop
         Last := Last + 1;
      end loop;

      Directories.Obj_Dir_Index := Last + 1;
      Directories.Last_Obj_Dir_Start := First;
      return Directories.Obj_Dir (First .. Last - 1);
   end Next_Obj_Dir;

   -------------------------
   -- Next_Unvisited_File --
   -------------------------

   function Next_Unvisited_File return File_Reference is
      The_Files : File_Reference := Files.Table;

   begin
      while The_Files /= null loop
         if not The_Files.Visited then
            The_Files.Visited := True;
            return The_Files;
         end if;

         The_Files := The_Files.Next;
      end loop;

      return Empty_File;
   end Next_Unvisited_File;

   ------------------
   -- Parse_Gnatls --
   ------------------

   procedure Parse_Gnatls
     (Include_Dir      : in  String;
      Gnatls_Src_Cache : out Ada.Strings.Unbounded.Unbounded_String;
      Gnatls_Obj_Cache : out Ada.Strings.Unbounded.Unbounded_String)
   is
      use type System.Address;

      function popen
        (Command   : System.Address;
         File_Type : System.Address)
         return      Interfaces.C_Streams.FILEs;
      pragma Import (C, popen, "popen");

      function pclose
        (File : Interfaces.C_Streams.FILEs)
         return Interfaces.C.int;
      pragma Import (C, pclose, "pclose");

      function Get_Directory (X : in String) return String;
      --  Clean a c-format string to get a proper Ada-format string

      -------------------
      -- Get_Directory --
      -------------------

      function Get_Directory (X : in String) return String is
         First, Last : Natural;

      begin
         First := X'First;
         while X (First) = ' ' loop
            First := First + 1;
         end loop;

         Last := First;
         while X (Last) /= ' '       and then
               X (Last) /= ASCII.NUL and then
               X (Last) /= ASCII.LF
         loop
            Last := Last + 1;
         end loop;

         Last := Last - 1;
         return X (First .. Last);

      exception
         when Constraint_Error => return "";
      end Get_Directory;


      File_Ptr  : Interfaces.C_Streams.FILEs;
      Results   : Line_String;
      Command   : constant String := "gnatls -v " & Include_Dir & ASCII.NUL;
      File_Type : constant String := "r" & ASCII.NUL;
      Done      : Interfaces.C_Streams.chars;
      Discard   : Interfaces.C.int;

   begin
      File_Ptr := popen (Command'Address, File_Type'Address);

      if File_Ptr = System.Null_Address then
         return;
      end if;

      --  Get lines from gnatls until "Source Search Path"

      loop
         Done := Interfaces.C_Streams.fgets (Results'Address,
                                             Results'Length,
                                             File_Ptr);
         exit when Results (1 .. 18) = "Source Search Path";
      end loop;

      --  Parse Src directories

      loop
         Done := Interfaces.C_Streams.fgets (Results'Address,
                                             Results'Length,
                                             File_Ptr);
         exit when Results (1 .. 18) = "Object Search Path";

         declare
            Directory : constant String := Get_Directory (Results);

         begin
            if Directory'Length > 0 then
               if Directory /= "<Current_Directory>" then
                  Ada.Strings.Unbounded.Append
                    (Gnatls_Src_Cache, Directory & ' ');
               else
                  Ada.Strings.Unbounded.Append (Gnatls_Src_Cache, "./ ");
               end if;
            end if;
         end;
      end loop;

      --  Parse Object directories
      loop
         Done := Interfaces.C_Streams.fgets (Results'Address,
                                             Results'Length,
                                             File_Ptr);
         declare
            Directory : constant String := Get_Directory (Results);

         begin
            exit when Directory'Length = 0;
            if Directory /= "<Current_Directory>" then
               Ada.Strings.Unbounded.Append
                 (Gnatls_Obj_Cache, Directory & ' ');
            else
               Ada.Strings.Unbounded.Append (Gnatls_Obj_Cache, "./ ");
            end if;
         end;
      end loop;

      Discard := pclose (File_Ptr);
   end Parse_Gnatls;

   -------------------
   -- Reset_Src_Dir --
   -------------------

   procedure Reset_Src_Dir is
   begin
      Directories.Src_Dir_Index := 1;
   end Reset_Src_Dir;

   -------------------
   -- Reset_Obj_Dir --
   -------------------

   procedure Reset_Obj_Dir is
   begin
      Directories.Obj_Dir_Index := 1;
   end Reset_Obj_Dir;

   -----------------------
   -- Set_Default_Match --
   -----------------------

   procedure Set_Default_Match (Value : Boolean) is
   begin
      Default_Match := Value;
   end Set_Default_Match;

   -------------------
   -- Set_Directory --
   -------------------

   procedure Set_Directory
     (File : in File_Reference;
      Dir  : in String)
   is
   begin
      File.Dir := new String'(Dir);
   end Set_Directory;

end Xr_Tabls;
