-----------------------------------------------------------------------------
--                                                                          --
--                         GNAT COMPILER COMPONENTS                         --
--                                                                          --
--                              X R E F _ L I B                             --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                            $Revision: 1.17 $                             --
--                                                                          --
--          Copyright (C) 1998 Free Software Foundation, Inc.               --
--                                                                          --
-- GNAT is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  GNAT is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNAT;  see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- GNAT is maintained by Ada Core Technologies Inc (http://www.gnat.com).   --
--                                                                          --
------------------------------------------------------------------------------

with Interfaces.C;
with Ada.Strings.Fixed;
with Ada.Strings.Maps.Constants;
with Ada.Strings.Unbounded;
with GNAT.OS_Lib;
with GNAT.IO_Aux;
with Ada.Characters.Latin_1;

package body Xref_Lib is

   use type Interfaces.C.size_t;
   use type System.Address;
   use type Ada.Strings.Maps.Character_Set;

   ---------------------
   -- Local Variables --
   ---------------------

   D    : constant Character := 'D';
   X    : constant Character := 'X';
   W    : constant Character := 'W';
   Dot  : constant Character := '.';
   Pipe : constant Character := '|';
   --  First character on xref lines in the .ali file

   No_Xref_Information : exception;
   --  Exception raised when there is no cross-referencing information in
   --  the .ali files

   ----------------
   -- Add_Entity --
   ----------------

   procedure Add_Entity
     (Pattern : in out Search_Pattern;
      Entity  : in String;
      Glob    : in Boolean := False)
   is
      File_Start   : Natural;
      Line_Start   : Natural;
      Col_Start    : Natural;
      Line_Num     : Natural := 0;
      Col_Num      : Natural := 0;
      File_Ref     : Xr_Tabls.File_Reference := Xr_Tabls.Empty_File;
      File_Existed : Boolean;
      Has_Pattern  : Boolean := False;

   begin
      --  Find the end of the first item in Entity (pattern or file?)
      --  If there is no ':', we only have a pattern

      File_Start := Ada.Strings.Fixed.Index (Entity, ":");

      if File_Start = 0 then
         Pattern.Entity := GNAT.Regexp.Compile (Entity, Glob, False);
         Xr_Tabls.Set_Default_Match (True);
         return;
      end if;

      --  If there is a dot in the pattern, then it is a file name

      if (Glob
          and then Ada.Strings.Fixed.Index
                     (Entity (Entity'First .. File_Start - 1), ".") /= 0)
        or else
          (not Glob
             and then Ada.Strings.Fixed.Index
                       (Entity (Entity'First .. File_Start - 1), "\.") /= 0)
      then
         Pattern.Entity := GNAT.Regexp.Compile (".*", False);
         File_Start     := Entity'First;

      else
         Pattern.Entity := GNAT.Regexp.Compile
           (Entity (Entity'First .. File_Start - 1), Glob, False);
         File_Start     := File_Start + 1;
         Has_Pattern    := True;
      end if;

      --  Parse the file name

      Line_Start := Ada.Strings.Fixed.Index
        (Entity (File_Start .. Entity'Last), ":");

      if Line_Start = 0 then
         Line_Start := Entity'Length + 1;

      elsif Line_Start /= Entity'Length then
         Col_Start := Ada.Strings.Fixed.Index
           (Entity (Line_Start + 1 .. Entity'Last), ":");

         if Col_Start = 0 then
            Col_Start := Entity'Last + 1;
         end if;

         if Col_Start > Line_Start + 2 then
            begin
               Line_Num := Natural'Value
                 (Entity (Line_Start + 1 .. Col_Start - 1));
            exception
               when Constraint_Error =>
                  raise Invalid_Argument;
            end;
         end if;

         if Col_Start < Entity'Last then
            begin
               Col_Num := Natural'Value (Entity
                                         (Col_Start + 1 .. Entity'Last));
            exception
               when Constraint_Error =>
                  raise Invalid_Argument;
            end;
         end if;
      end if;

      Xr_Tabls.Add_File (Entity (File_Start .. Line_Start - 1),
                         File_Existed,
                         File_Ref,
                         Visited => True);
      Xr_Tabls.Add_Line (File_Ref, Line_Num, Col_Num);
      Xr_Tabls.Add_File
        (Xr_Tabls.ALI_File_Name (Entity (File_Start .. Line_Start - 1)),
         File_Existed, File_Ref,
         Visited => False,
         Emit_Warning => True);
   end Add_Entity;

   --------------
   -- Add_File --
   --------------

   procedure Add_File (File : String) is
      File_Ref     : Xr_Tabls.File_Reference := Xr_Tabls.Empty_File;
      File_Existed : Boolean;

   begin
      --  Case where we have an ALI file, accept it even though this is
      --  not official usage, since the intention is obvious

      if Ada.Strings.Fixed.Tail (File, 3) = "ali" then
         Xr_Tabls.Add_File (File, File_Existed, File_Ref, Visited => False,
                            Emit_Warning => True);

      --  Normal non-ali file case

      else
         Xr_Tabls.Add_File (File, File_Existed, File_Ref, Visited => True);
         Xr_Tabls.Add_File (Xr_Tabls.ALI_File_Name (File),
                            File_Existed, File_Ref, Visited => False,
                            Emit_Warning => True);
      end if;
   end Add_File;

   -----------
   -- Close --
   -----------

   procedure Close (File : in out ALI_File) is
   begin
      Ada.Text_IO.Close (File.File);
   end Close;

   -----------------------
   -- Current_Xref_File --
   -----------------------

   function Current_Xref_File
     (File : ALI_File)
      return Xr_Tabls.File_Reference
   is
   begin
      return File.Current_File;
   end Current_Xref_File;

   --------------------------
   -- Default_Project_File --
   --------------------------

   function Default_Project_File
     (Dir_Name : String)
      return     String
   is
      use type GNAT.OS_Lib.String_Access;

      My_Dir  : GNAT.Directory_Operations.Dir_Type;
      Dir_Ent : File_Name_String;
      Last    : Natural;

   begin
      GNAT.Directory_Operations.Open (My_Dir, Dir_Name);

      loop
         GNAT.Directory_Operations.Read (My_Dir, Dir_Ent, Last);
         exit when Last = 0;

         if Ada.Strings.Fixed.Tail (Dir_Ent (1 .. Last), 4) = ".adp" then
            --  The first project file found is the good one.
            GNAT.Directory_Operations.Close (My_Dir);
            return Dir_Ent (1 .. Last);
         end if;
      end loop;

      GNAT.Directory_Operations.Close (My_Dir);
      return String'(1 .. 0 => ' ');

   exception
      when GNAT.Directory_Operations.Directory_Error =>
         return String'(1 .. 0 => ' ');
   end Default_Project_File;

   ---------------
   -- File_Name --
   ---------------

   function File_Name
     (File : ALI_File;
      Num  : Positive)
      return Xr_Tabls.File_Reference
   is
   begin
      return File.Dep (Num);
   end File_Name;

   --------------------
   -- Find_ALI_Files --
   --------------------

   procedure Find_ALI_Files is
      My_Dir       : Rec_DIR;
      Dir_Ent      : File_Name_String;
      Last         : Natural;
      File_Existed : Boolean;
      File_Ref     : Xr_Tabls.File_Reference;

      function Open_Next_Dir return Boolean;
      --  Tries to open the next object directory, and return False if
      --  the directory cannot be opened.

      -------------------
      -- Open_Next_Dir --
      -------------------

      function Open_Next_Dir return Boolean is
      begin
         --  Until we are able to open a new directory

         loop
            declare
               Obj_Dir : constant String := Xr_Tabls.Next_Obj_Dir;
            begin
               --  If there was no more Obj_Dir line

               if Obj_Dir'Length = 0 then
                  return False;
               end if;

               GNAT.Directory_Operations.Open (My_Dir.Dir, Obj_Dir);
               exit;

            exception
               --  Could not open the directory
               when others =>
                  null;
            end;
         end loop;

         return True;
      end Open_Next_Dir;

   --  Start of processing for Find_ALI_Files

   begin
      if Open_Next_Dir then
         loop
            GNAT.Directory_Operations.Read (My_Dir.Dir, Dir_Ent, Last);

            if Last = 0 then
               GNAT.Directory_Operations.Close (My_Dir.Dir);

               if not Open_Next_Dir then
                  return;
               end if;

            elsif Ada.Strings.Fixed.Tail (Dir_Ent (1 .. Last), 4) = ".ali" then
               Xr_Tabls.Add_File (Dir_Ent (1 .. Last), File_Existed,
                                  File_Ref, Visited => False);
               Xr_Tabls.Set_Directory (File_Ref, Xr_Tabls.Current_Obj_Dir);
            end if;
         end loop;
      end if;
   end Find_ALI_Files;

   ------------------
   -- Symbol_Match --
   ------------------

   function Match
     (Pattern : Search_Pattern;
      Symbol  : String)
      return    Boolean
   is
      use type Xr_Tabls.File_Reference;

   begin
      --  Get the entity name

      return GNAT.Regexp.Match (Symbol, Pattern.Entity);
   end Match;

   ---------------------
   -- Next_Identifier --
   ---------------------

   procedure Next_Identifier
     (File : in out ALI_File;
      Line : out Ada.Strings.Unbounded.Unbounded_String)
   is
      First : Positive;

   begin
      Line := Ada.Strings.Unbounded.Null_Unbounded_String;

      if File.Next_Line_Length = 0 then
         return;
      end if;

      while not Ada.Text_IO.End_Of_File (File.File) loop

         case File.Next_Line (1) is

            when X =>

               --  We had a line giving a new file

               First := Ada.Strings.Fixed.Index
                 (File.Next_Line (3 .. File.Next_Line_Length), " ");
               File.Current_File :=
                 File.Dep (Positive'Value (File.Next_Line
                                              (3 .. First - 1)));

            when Dot =>
               Ada.Strings.Unbounded.Append
                 (Line, File.Next_Line (2 .. File.Next_Line_Length));

            when others =>
               Line := Ada.Strings.Unbounded.To_Unbounded_String
                 (File.Next_Line (1 .. File.Next_Line_Length));
         end case;

         Ada.Text_IO.Get_Line
           (File.File, File.Next_Line, File.Next_Line_Length);

         --  If the next line as nothing to do with the current one then

         if File.Next_Line_Length > 1
           and then File.Next_Line (1) /= Dot
           and then Ada.Strings.Unbounded.Length (Line) /= 0
         then
            return;
         end if;
      end loop;

      if File.Next_Line (1) /= Dot then
         Line := Ada.Strings.Unbounded.To_Unbounded_String
           (File.Next_Line (1 .. File.Next_Line_Length));
      else
         Ada.Strings.Unbounded.Append
           (Line, File.Next_Line (2 .. File.Next_Line_Length));
      end if;
      File.Next_Line_Length := 0;
   end Next_Identifier;

   ----------
   -- Open --
   ----------

   procedure Open
     (Name         : String;
      File         : out ALI_File;
      Dependencies : Boolean := False)
   is
      Line             : Line_String;
      Length           : Natural;
      First, Last      : Integer;
      Num_Dependencies : Natural := 0;
      File_Existed     : Boolean;
      File_Ref         : Xr_Tabls.File_Reference;

   begin
      Ada.Text_IO.Open (File => File.File,
                        Mode => Ada.Text_IO.In_File,
                        Name => Name);

      --  Ignore all the lines until the first Xref line

      while not Ada.Text_IO.End_Of_File (File.File) loop

         Ada.Text_IO.Get_Line (File => File.File,
                               Item => Line,
                               Last => Length);

         --  Case where empty line was not read

         if Length >= Line'First then

            --  If we have read the beginning of the xref section

            if Line (Line'First) = X then
               File.Next_Line := Line;
               File.Next_Line_Length := Length;
               return;

            elsif Line (Line'First) = D then
               Num_Dependencies := Num_Dependencies + 1;
               First := 3;

               while First <= Line'Last
                 and then Line (First) /= ' '
                 and then Line (First) /= Ada.Characters.Latin_1.HT
               loop
                  First := First + 1;
               end loop;

               Xr_Tabls.Add_File (Line (3 .. First - 1),
                                  File_Existed,
                                  File.Dep (Num_Dependencies));

            --  In gnatxref mode, we want to keep trace of the dependencies

            elsif Dependencies and then Line (Line'First) = W then

               --  Find the .ali file name

               Last := Length;
               while Last > Line'First + 2
                 and then Line (Last - 2 .. Last) /= "ali"
               loop
                  Last := Last - 1;
               end loop;

               if Last > Line'First + 2 then

                  --  Find the beginning of the name

                  First := Last - 2;

                  while Line (First) /= ' '
                    and Line (First) /= Ada.Characters.Latin_1.HT
                  loop
                     First := First - 1;
                  end loop;

                  Xr_Tabls.Add_File (Line (First + 1 .. Last),
                                     File_Existed, File_Ref,
                                     Visited => False);
               end if;
            end if;
         end if;
      end loop;

      --  No xref information in the .ali file

      File.Next_Line_Length := 0;
      raise No_Xref_Information;
   end Open;

   --------------------
   -- Print_Gnatfind --
   --------------------

   procedure Print_Gnatfind
     (References     : in Boolean;
      Full_Path_Name : in Boolean)
   is
      use type Xr_Tabls.Declaration_Reference;
      use type Xr_Tabls.Reference;
      use type Xr_Tabls.Compare_Result;

      Decl : Xr_Tabls.Declaration_Reference := Xr_Tabls.First_Declaration;
      Ref1 : Xr_Tabls.Reference;
      Ref2 : Xr_Tabls.Reference;

      procedure Print_Ref
        (Ref : Xr_Tabls.Reference;
         Msg : String := "");
      --  Print a reference, according to the extended tag of the output

      ---------------
      -- Print_Ref --
      ---------------

      procedure Print_Ref
        (Ref : Xr_Tabls.Reference;
         Msg : String := "")
      is
      begin
         Ada.Text_IO.Put_Line
           (Xr_Tabls.Get_File (Ref, Full_Path_Name)
            & ":" & Xr_Tabls.Get_Line (Ref)
            & ":" & Xr_Tabls.Get_Column (Ref)
            & ":  " & Xr_Tabls.Get_Symbol (Decl)
            & Msg);
      end Print_Ref;

   --  Start of processing for Print_Gnatfind

   begin
      while Decl /= Xr_Tabls.Empty_Declaration loop
         if Xr_Tabls.Match (Decl) then

            --  Output the declaration

            Ada.Text_IO.Put_Line
              (Xr_Tabls.Get_File (Decl, Full_Path_Name)
                & ":" & Xr_Tabls.Get_Line (Decl)
                & ":" & Xr_Tabls.Get_Column (Decl)
                & ":  " & Xr_Tabls.Get_Symbol (Decl)
                & "   <= declaration");

            --  Output the body (sorted)

            Ref1 := Xr_Tabls.First_Body (Decl);
            while Ref1 /= Xr_Tabls.Empty_Reference loop
               Print_Ref (Ref1, "   <= body");
               Ref1 := Xr_Tabls.Next (Ref1);
            end loop;

            if References then
               Ref1 := Xr_Tabls.First_Modif (Decl);
               Ref2 := Xr_Tabls.First_Reference (Decl);
               while Ref1 /= Xr_Tabls.Empty_Reference
                       or else
                     Ref2 /= Xr_Tabls.Empty_Reference
               loop
                  if Xr_Tabls.Compare (Ref1, Ref2) = Xr_Tabls.LessThan then
                     Print_Ref (Ref1);
                     Ref1 := Xr_Tabls.Next (Ref1);
                  else
                     Print_Ref (Ref2);
                     Ref2 := Xr_Tabls.Next (Ref2);
                  end if;
               end loop;
            end if;
         end if;

         Decl := Xr_Tabls.Next (Decl);
      end loop;
   end Print_Gnatfind;

   --------------
   -- Print_Vi --
   --------------

   procedure Print_Vi (Full_Path_Name : in Boolean) is
      use type Xr_Tabls.Declaration_Reference;
      use type Xr_Tabls.Reference;
      use type Xr_Tabls.File_Reference;

      Tab  : constant Character := Ada.Characters.Latin_1.HT;
      Decl : Xr_Tabls.Declaration_Reference := Xr_Tabls.First_Declaration;
      Ref  : Xr_Tabls.Reference;

   begin
      while Decl /= Xr_Tabls.Empty_Declaration loop
         Ada.Text_IO.Put_Line (Xr_Tabls.Get_Symbol (Decl) & Tab
                               & Xr_Tabls.Get_File (Decl, Full_Path_Name) & Tab
                               & Xr_Tabls.Get_Line (Decl));

         --  Print the body if any

         Ref := Xr_Tabls.First_Body (Decl);

         if Ref /= Xr_Tabls.Empty_Reference then
            Ada.Text_IO.Put_Line (Xr_Tabls.Get_Symbol (Decl) & Tab
                                  & Xr_Tabls.Get_File (Ref, Full_Path_Name)
                                  & Tab
                                  & Xr_Tabls.Get_Line (Ref));
         end if;

         --  Print the modifications

         Ref := Xr_Tabls.First_Modif (Decl);

         while Ref /= Xr_Tabls.Empty_Reference loop
            Ada.Text_IO.Put_Line (Xr_Tabls.Get_Symbol (Decl) & Tab
                                  & Xr_Tabls.Get_File (Ref, Full_Path_Name)
                                  & Tab
                                  & Xr_Tabls.Get_Line (Ref));
            Ref := Xr_Tabls.Next (Ref);
         end loop;

         Decl := Xr_Tabls.Next (Decl);
      end loop;
   end Print_Vi;

   ----------------
   -- Print_Xref --
   ----------------

   procedure Print_Xref (Full_Path_Name : in Boolean) is
      use type Xr_Tabls.Declaration_Reference;
      use type Xr_Tabls.Reference;
      use type Xr_Tabls.File_Reference;

      Decl : Xr_Tabls.Declaration_Reference := Xr_Tabls.First_Declaration;
      Ref  : Xr_Tabls.Reference;
      File : Xr_Tabls.File_Reference;
      Col  : Positive := 1;

      procedure New_Line80;
      procedure Print80 (S : in String);
      --  Print the text, respecting the 80 columns rule

      procedure New_Line80 is
      begin
         Ada.Text_IO.New_Line;
         Col := 1;
      end New_Line80;

      procedure Print80 (S : in String) is
      begin
         if Col + S'Length > 79 then
            Ada.Text_IO.New_Line;
            Ada.Text_IO.Put ("  ");
            Col := 2;
         end if;
         Ada.Text_IO.Put (S);
         Col := Col + S'Length;
      end Print80;

   --  Start of processing for Print_Xref

   begin
      while Decl /= Xr_Tabls.Empty_Declaration loop
         Print80 (Xr_Tabls.Get_Symbol (Decl)
                   & " " & Xr_Tabls.Get_Type (Decl)
                   & " ");
         Print80 (Xr_Tabls.Get_File (Decl, Full_Path_Name) & ':'
                   & Xr_Tabls.Get_Line (Decl) & ':'
                   & Xr_Tabls.Get_Column (Decl));

         --  Print the body if any

         Ref := Xr_Tabls.First_Body (Decl);

         if Ref /= Xr_Tabls.Empty_Reference then
            Print80 (' '
                      & Xr_Tabls.Get_File (Ref, Full_Path_Name)
                      & ':' & Xr_Tabls.Get_Line (Ref)
                      & ':' & Xr_Tabls.Get_Column (Ref));
         end if;

         --  Print the modifications

         Print80 (" {");
         Ref := Xr_Tabls.First_Modif (Decl);
         File := Xr_Tabls.Empty_File;

         while Ref /= Xr_Tabls.Empty_Reference loop
            if Xr_Tabls.Get_File_Ref (Ref) /= File then
               File := Xr_Tabls.Get_File_Ref (Ref);
               Print80 (Xr_Tabls.Get_File (Ref, Full_Path_Name) & ':'
                         & Xr_Tabls.Get_Line (Ref)
                         & ':' & Xr_Tabls.Get_Column (Ref)
                         & ' ');
            else
               Print80 (Xr_Tabls.Get_Line (Ref)
                         & ':' & Xr_Tabls.Get_Column (Ref)
                         & ' ');
            end if;

            Ref := Xr_Tabls.Next (Ref);
         end loop;

         Print80 ("} {");

         --  Print the references

         Ref := Xr_Tabls.First_Reference (Decl);
         File := Xr_Tabls.Empty_File;

         while Ref /= Xr_Tabls.Empty_Reference loop
            if Xr_Tabls.Get_File_Ref (Ref) /= File then
               File := Xr_Tabls.Get_File_Ref (Ref);
               Print80 (Xr_Tabls.Get_File (Ref, Full_Path_Name) & ':'
                         & Xr_Tabls.Get_Line (Ref)
                         & ':' & Xr_Tabls.Get_Column (Ref)
                         & ' ');
            else
               Print80 (Xr_Tabls.Get_Line (Ref)
                         & ':' & Xr_Tabls.Get_Column (Ref)
                         & ' ');
            end if;

            Ref := Xr_Tabls.Next (Ref);
         end loop;

         Print80 ("}");
         New_Line80;
         Decl := Xr_Tabls.Next (Decl);
      end loop;
   end Print_Xref;

   ------------------
   -- Print_Unused --
   ------------------

   procedure Print_Unused (Full_Path_Name : in Boolean) is
      use Xr_Tabls;

      Decl : Xr_Tabls.Declaration_Reference := Xr_Tabls.First_Declaration;
      Ref  : Xr_Tabls.Reference;

   begin
      while Decl /= Empty_Declaration loop
         if First_Modif (Decl) = Empty_Reference
           and then First_Reference (Decl) = Empty_Reference
         then
            Ada.Text_IO.Put (Get_Symbol (Decl)
                              & " "
                              & Get_Type (Decl)
                              & " "
                              & Get_File (Decl, Full_Path_Name)
                              & ':'
                              & Get_Line (Decl)
                              & ':'
                              & Get_Column (Decl));

            --  Print the body if any

            Ref := First_Body (Decl);

            if Ref /= Empty_Reference then
               Ada.Text_IO.Put_Line (' '
                                      & Get_File (Ref, Full_Path_Name)
                                      & ':' & Get_Line (Ref)
                                      & ':' & Get_Column (Ref));
            else
               Ada.Text_IO.New_Line;
            end if;
         end if;

         Decl := Next (Decl);
      end loop;
   end Print_Unused;

   -------------
   -- Process --
   -------------

   procedure Process
     (Pattern       : Search_Pattern;
      Line          : String;
      ALIfile       : ALI_File;
      Local_Symbols : Boolean)
   is
      use Ada.Strings.Maps.Constants;
      use type Xr_Tabls.Declaration_Reference;

      Symbol_Last : Natural;

      First    : Natural;
      Last     : Natural;
      Ref_Type : Natural;
      Decl_Ref : Xr_Tabls.Declaration_Reference;
      File_Ref : Xr_Tabls.File_Reference := Current_Xref_File (ALIfile);

   begin
      --  Extract the line, column and entity name informations

      First := Ada.Strings.Fixed.Index (Line, not Decimal_Digit_Set);
      Last  := Ada.Strings.Fixed.Index (Line (First + 1 .. Line'Last),
                                        not Decimal_Digit_Set);
      Symbol_Last := Ada.Strings.Fixed.Index (Line (Last + 1 .. Line'Last),
                                              " ");
      if Symbol_Last = 0 then
         Symbol_Last := Line'Last + 1;
      end if;

      --  Exit if the symbol does not match
      --  or if we have a local symbol and we do not want them

      if not Match (Pattern, Line (Last + 1 .. Symbol_Last - 1))
        or else
        (Line (Last) = ' ' and then not Local_Symbols)
      then
         return;
      end if;

      --  Insert the declaration in the table

      Decl_Ref := Xr_Tabls.Add_Declaration
        (Current_Xref_File (ALIfile),                    --  File
         Line (Last + 1 .. Symbol_Last - 1),             --  Entity name
         Natural'Value (Line (Line'First .. First - 1)), --  Line
         Natural'Value (Line (First + 1 .. Last - 1)),   --  Column
         Line (First));                                  --  Type

      --  To find the body, we will have to parse the file too

      declare
         File_Existed : Boolean;
         File_Ref     : Xr_Tabls.File_Reference;
         File_Name    : String := Xr_Tabls.Get_File
           (Current_Xref_File (ALIfile));
      begin
         Xr_Tabls.Add_File (Xr_Tabls.ALI_File_Name (File_Name), File_Existed,
                            File_Ref, False);
      end;

      --  For every item on the line

      First := Symbol_Last + 1;
      loop
         Ref_Type := Ada.Strings.Fixed.Index (Line (First .. Line'Last),
                                              not Decimal_Digit_Set);
         exit when Ref_Type = 0;

         if Line (Ref_Type) = Pipe then
            File_Ref := File_Name
              (ALIfile, Integer'Value (Line (First .. Ref_Type - 1)));
            First := Ref_Type + 1;
            Ref_Type := Ada.Strings.Fixed.Index
              (Line (First .. Line'Last), not Decimal_Digit_Set);
         end if;

         Last := Ada.Strings.Fixed.Index (Line (Ref_Type + 1 .. Line'Last),
                                          not Decimal_Digit_Set);
         if Last = 0 then
            Last := Line'Last + 1;
         end if;

         --  Insert the reference or body in the table

         Xr_Tabls.Add_Reference
           (Decl_Ref, File_Ref,
            Integer'Value (Line (First .. Ref_Type - 1)),
            Integer'Value (Line (Ref_Type + 1 .. Last - 1)),
            Line (Ref_Type));

         --  Next iteration

         First := Last + 1;
      end loop;
   end Process;

   ------------
   -- Search --
   ------------

   procedure Search
     (Pattern       : Search_Pattern;
      Local_Symbols : Boolean;
      Wide_Search   : Boolean;
      Read_Only     : Boolean)
   is
      use Ada.Strings.Unbounded;
      use type Xr_Tabls.File_Reference;

      ALIfile               : ALI_File;
      Line                  : Unbounded_String;
      File_Ref              : Xr_Tabls.File_Reference;

   begin
      --  If we want all the .ali files, then find them

      if Wide_Search then
         Find_ALI_Files;
      end if;

      loop
         --  Get the next unread ali file

         File_Ref := Xr_Tabls.Next_Unvisited_File;
         exit when File_Ref = Xr_Tabls.Empty_File;

         if not GNAT.IO_Aux.File_Exists
           (Xr_Tabls.Get_File (File_Ref, With_Dir => True))
         then

            if Xr_Tabls.Get_Emit_Warning (File_Ref) then
               Ada.Text_IO.Put_Line
                 (Ada.Text_IO.Standard_Error,
                  "warning : file "
                  & Xr_Tabls.Get_File (File_Ref, With_Dir => True)
                  & " not found");
            end if;

         else
            --  Case of read only file

            if Read_Only
              or else GNAT.OS_Lib.Is_Writable_File
                        (Xr_Tabls.Get_File (File_Ref, With_Dir => True))
            then
               begin
                  Open (Xr_Tabls.Get_File (File_Ref, With_Dir => True),
                        ALIfile);
                  loop
                     Next_Identifier (ALIfile, Line);
                     exit when Ada.Strings.Unbounded.Length (Line) = 0;
                     Process (Pattern, To_String (Line), ALIfile,
                              Local_Symbols);
                  end loop;

                  Close (ALIfile);

               exception
                  when No_Xref_Information   =>
                     if Xr_Tabls.Get_Emit_Warning (File_Ref) then
                        Ada.Text_IO.Put_Line
                          (Ada.Text_IO.Standard_Error,
                           "warning : No cross-referencing information in  "
                           & Xr_Tabls.Get_File (File_Ref, With_Dir => True));
                     end if;
                     Close (ALIfile);

                  when Ada.Text_IO.Status_Error =>
                     Ada.Text_IO.Put_Line ("Status Error on "
                                           & Xr_Tabls.Get_File (File_Ref));
                  when others =>
                     null;
               end;
            end if;
         end if;
      end loop;
   end Search;

   -----------------
   -- Search_Xref --
   -----------------

   procedure Search_Xref
     (Pattern       : Search_Pattern;
      Local_Symbols : Boolean;
      Read_Only     : Boolean)
   is
      use Ada.Strings.Unbounded;
      use type Xr_Tabls.File_Reference;

      ALIfile     : ALI_File;
      Line        : Unbounded_String;
      File_Ref    : Xr_Tabls.File_Reference;
   begin
      loop
         --  Find the next unvisited file

         File_Ref := Xr_Tabls.Next_Unvisited_File;
         exit when File_Ref = Xr_Tabls.Empty_File;

         --  Search the object directories for the .ali file

         if Read_Only
           or else GNAT.OS_Lib.Is_Writable_File
                     (Xr_Tabls.Get_File (File_Ref, With_Dir => True))
         then
            begin
               Open (Xr_Tabls.Get_File (File_Ref, With_Dir => True),
                     ALIfile, True);

               loop
                  Next_Identifier (ALIfile, Line);
                  exit when Ada.Strings.Unbounded.Length (Line) = 0;
                  Process (Pattern, To_String (Line), ALIfile,
                           Local_Symbols);
               end loop;

               Close (ALIfile);

            exception
               when Ada.Text_IO.Name_Error =>
                  null;
            end;
         end if;
      end loop;
   end Search_Xref;

end Xref_Lib;
